#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine.resources import stack_user

LOG = logging.getLogger(__name__)

#
# We are ignoring Groups as keystone does not support them.
# For now support users and accesskeys,
# We also now support a limited heat-native Policy implementation, and
# the native access policy resource is located at:
# heat/engine/resources/openstack/access_policy.py
#


class User(stack_user.StackUser):
    PROPERTIES = (
        PATH, GROUPS, LOGIN_PROFILE, POLICIES,
    ) = (
        'Path', 'Groups', 'LoginProfile', 'Policies',
    )

    _LOGIN_PROFILE_KEYS = (
        LOGIN_PROFILE_PASSWORD,
    ) = (
        'Password',
    )

    properties_schema = {
        PATH: properties.Schema(
            properties.Schema.STRING,
            _('Not Implemented.')
        ),
        GROUPS: properties.Schema(
            properties.Schema.LIST,
            _('Not Implemented.')
        ),
        LOGIN_PROFILE: properties.Schema(
            properties.Schema.MAP,
            _('A login profile for the user.'),
            schema={
                LOGIN_PROFILE_PASSWORD: properties.Schema(
                    properties.Schema.STRING
                ),
            }
        ),
        POLICIES: properties.Schema(
            properties.Schema.LIST,
            _('Access policies to apply to the user.')
        ),
    }

    def _validate_policies(self, policies):
        for policy in (policies or []):
            # When we support AWS IAM style policies, we will have to accept
            # either a ref to an AWS::IAM::Policy defined in the stack, or
            # and embedded dict describing the policy directly, but for now
            # we only expect this list to contain strings, which must map
            # to an OS::Heat::AccessPolicy in this stack
            # If a non-string (e.g embedded IAM dict policy) is passed, we
            # ignore the policy (don't reject it because we previously ignored
            # and we don't want to break templates which previously worked
            if not isinstance(policy, str):
                LOG.debug("Ignoring policy %s, must be string "
                          "resource name", policy)
                continue

            try:
                policy_rsrc = self.stack[policy]
            except KeyError:
                LOG.debug("Policy %(policy)s does not exist in stack "
                          "%(stack)s",
                          {'policy': policy, 'stack': self.stack.name})
                return False

            if not callable(getattr(policy_rsrc, 'access_allowed', None)):
                LOG.debug("Policy %s is not an AccessPolicy resource", policy)
                return False

        return True

    def handle_create(self):
        profile = self.properties[self.LOGIN_PROFILE]
        if profile and self.LOGIN_PROFILE_PASSWORD in profile:
            self.password = profile[self.LOGIN_PROFILE_PASSWORD]

        if self.properties[self.POLICIES]:
            if not self._validate_policies(self.properties[self.POLICIES]):
                raise exception.InvalidTemplateAttribute(resource=self.name,
                                                         key=self.POLICIES)

        super(User, self).handle_create()
        self.resource_id_set(self._get_user_id())

    def get_reference_id(self):
        return self.physical_resource_name_or_FnGetRefId()

    def access_allowed(self, resource_name):
        policies = (self.properties[self.POLICIES] or [])
        for policy in policies:
            if not isinstance(policy, str):
                LOG.debug("Ignoring policy %s, must be string "
                          "resource name", policy)
                continue
            policy_rsrc = self.stack[policy]
            if not policy_rsrc.access_allowed(resource_name):
                return False
        return True


class AccessKey(resource.Resource):
    PROPERTIES = (
        SERIAL, USER_NAME, STATUS,
    ) = (
        'Serial', 'UserName', 'Status',
    )

    ATTRIBUTES = (
        USER_NAME, SECRET_ACCESS_KEY,
    ) = (
        'UserName', 'SecretAccessKey',
    )

    properties_schema = {
        SERIAL: properties.Schema(
            properties.Schema.INTEGER,
            _('Not Implemented.'),
            implemented=False
        ),
        USER_NAME: properties.Schema(
            properties.Schema.STRING,
            _('The name of the user that the new key will belong to.'),
            required=True
        ),
        STATUS: properties.Schema(
            properties.Schema.STRING,
            _('Not Implemented.'),
            constraints=[
                constraints.AllowedValues(['Active', 'Inactive']),
            ],
            implemented=False
        ),
    }

    attributes_schema = {
        USER_NAME: attributes.Schema(
            _('Username associated with the AccessKey.'),
            cache_mode=attributes.Schema.CACHE_NONE,
            type=attributes.Schema.STRING
        ),
        SECRET_ACCESS_KEY: attributes.Schema(
            _('Keypair secret key.'),
            cache_mode=attributes.Schema.CACHE_NONE,
            type=attributes.Schema.STRING
        ),
    }

    def __init__(self, name, json_snippet, stack):
        super(AccessKey, self).__init__(name, json_snippet, stack)
        self._secret = None
        if self.resource_id:
            self._register_access_key()

    def _get_user(self):
        """Derive the keystone userid, stored in the User resource_id.

        Helper function to derive the keystone userid, which is stored in the
        resource_id of the User associated with this key. We want to avoid
        looking the name up via listing keystone users, as this requires admin
        rights in keystone, so FnGetAtt which calls _secret_accesskey won't
        work for normal non-admin users.
        """
        # Lookup User resource by intrinsic reference (which is what is passed
        # into the UserName parameter.  Would be cleaner to just make the User
        # resource return resource_id for FnGetRefId but the AWS definition of
        # user does say it returns a user name not ID
        return self.stack.resource_by_refid(self.properties[self.USER_NAME])

    def handle_create(self):
        user = self._get_user()
        if user is None:
            raise exception.NotFound(_('could not find user %s') %
                                     self.properties[self.USER_NAME])
        # The keypair is actually created and owned by the User resource
        kp = user._create_keypair()
        self.resource_id_set(kp.access)
        self._secret = kp.secret
        self._register_access_key()

        # Store the secret key, encrypted, in the DB so we don't have lookup
        # the user every time someone requests the SecretAccessKey attribute
        self.data_set('secret_key', kp.secret, redact=True)
        self.data_set('credential_id', kp.id, redact=True)

    def handle_delete(self):
        self._secret = None
        if self.resource_id is None:
            return

        user = self._get_user()
        if user is None:
            LOG.debug('Error deleting %s - user not found', str(self))
            return
        user._delete_keypair()

    def _secret_accesskey(self):
        """Return the user's access key.

        Fetching it from keystone if necessary.
        """
        if self._secret is None:
            if not self.resource_id:
                LOG.info('could not get secret for %(username)s '
                         'Error:%(msg)s',
                         {'username': self.properties[self.USER_NAME],
                          'msg': "resource_id not yet set"})
            else:
                # First try to retrieve the secret from resource_data, but
                # for backwards compatibility, fall back to requesting from
                # keystone
                self._secret = self.data().get('secret_key')
                if self._secret is None:
                    try:
                        user_id = self._get_user().resource_id
                        kp = self.keystone().get_ec2_keypair(
                            user_id=user_id, access=self.resource_id)
                        self._secret = kp.secret
                        # Store the key in resource_data
                        self.data_set('secret_key', kp.secret, redact=True)
                        # And the ID of the v3 credential
                        self.data_set('credential_id', kp.id, redact=True)
                    except Exception as ex:
                        LOG.info('could not get secret for %(username)s '
                                 'Error:%(msg)s',
                                 {'username': self.properties[self.USER_NAME],
                                  'msg': ex})

        return self._secret or '000-000-000'

    def _resolve_attribute(self, name):
        if name == self.USER_NAME:
            return self.properties[self.USER_NAME]
        elif name == self.SECRET_ACCESS_KEY:
            return self._secret_accesskey()

    def _register_access_key(self):

        def access_allowed(resource_name):
            return self._get_user().access_allowed(resource_name)
        self.stack.register_access_allowed_handler(
            self.resource_id, access_allowed)


def resource_mapping():
    return {
        'AWS::IAM::User': User,
        'AWS::IAM::AccessKey': AccessKey,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import cinder as c_plugin
from heat.engine.clients.os import keystone as k_plugin
from heat.engine import rsrc_defn
from heat.engine import stack as parser
from heat.engine import template
from heat.tests import common
from heat.tests import utils

quota_template = '''
heat_template_version: newton

description: Sample cinder quota heat template

resources:
  my_quota:
    type: OS::Cinder::Quota
    properties:
      project: demo
      gigabytes: 5
      snapshots: 2
      volumes: 3
'''


class CinderQuotaTest(common.HeatTestCase):
    def setUp(self):
        super(CinderQuotaTest, self).setUp()

        self.ctx = utils.dummy_context()
        self.patchobject(c_plugin.CinderClientPlugin, 'has_extension',
                         return_value=True)
        self.patchobject(k_plugin.KeystoneClientPlugin, 'get_project_id',
                         return_value='some_project_id')
        tpl = template_format.parse(quota_template)
        self.stack = parser.Stack(
            self.ctx, 'cinder_quota_test_stack',
            template.Template(tpl)
        )

        self.my_quota = self.stack['my_quota']
        cinder = mock.MagicMock()
        self.cinderclient = mock.MagicMock()
        self.my_quota.client = cinder
        cinder.return_value = self.cinderclient
        self.quotas = self.cinderclient.quotas
        self.quota_set = mock.MagicMock()
        self.quotas.update.return_value = self.quota_set
        self.quotas.delete.return_value = self.quota_set

        class FakeVolumeOrSnapshot(object):
            def __init__(self, size=1):
                self.size = size
        self.fv = FakeVolumeOrSnapshot
        f_v = self.fv()
        self.volume_snapshots = self.cinderclient.volume_snapshots
        self.volume_snapshots.list.return_value = [f_v, f_v]
        self.volumes = self.cinderclient.volumes
        self.volumes.list.return_value = [f_v, f_v, f_v]
        self.err_msg = ("Invalid quota %(property)s value(s): %(value)s. "
                        "Can not be less than the current usage value(s): "
                        "%(total)s.")

    def _test_validate(self, resource, error_msg):
        exc = self.assertRaises(exception.StackValidationFailed,
                                resource.validate)
        self.assertIn(error_msg, str(exc))

    def _test_invalid_property(self, prop_name):
        my_quota = self.stack['my_quota']
        props = self.stack.t.t['resources']['my_quota']['properties'].copy()
        props[prop_name] = -2
        my_quota.t = my_quota.t.freeze(properties=props)
        my_quota.reparse()
        error_msg = ('Property error: resources.my_quota.properties.%s:'
                     ' -2 is out of range (min: -1, max: None)' % prop_name)
        self._test_validate(my_quota, error_msg)

    def test_invalid_gigabytes(self):
        self._test_invalid_property('gigabytes')

    def test_invalid_snapshots(self):
        self._test_invalid_property('snapshots')

    def test_invalid_volumes(self):
        self._test_invalid_property('volumes')

    def test_miss_all_quotas(self):
        my_quota = self.stack['my_quota']
        props = self.stack.t.t['resources']['my_quota']['properties'].copy()
        del props['gigabytes'], props['snapshots'], props['volumes']
        my_quota.t = my_quota.t.freeze(properties=props)
        my_quota.reparse()
        msg = ('At least one of the following properties must be specified: '
               'gigabytes, snapshots, volumes.')
        self.assertRaisesRegex(exception.PropertyUnspecifiedError, msg,
                               my_quota.validate)

    def test_quota_handle_create(self):
        self.my_quota.physical_resource_name = mock.MagicMock(
            return_value='some_resource_id')
        self.my_quota.reparse()
        self.my_quota.handle_create()
        self.quotas.update.assert_called_once_with(
            'some_project_id',
            gigabytes=5,
            snapshots=2,
            volumes=3
        )
        self.assertEqual('some_resource_id', self.my_quota.resource_id)

    def test_quota_handle_update(self):
        tmpl_diff = mock.MagicMock()
        prop_diff = mock.MagicMock()
        props = {'project': 'some_project_id', 'gigabytes': 6,
                 'volumes': 4}
        json_snippet = rsrc_defn.ResourceDefinition(
            self.my_quota.name,
            'OS::Cinder::Quota',
            properties=props)
        self.my_quota.reparse()
        self.my_quota.handle_update(json_snippet, tmpl_diff, prop_diff)
        self.quotas.update.assert_called_once_with(
            'some_project_id',
            gigabytes=6,
            volumes=4
        )

    def test_quota_handle_delete(self):
        self.my_quota.reparse()
        self.my_quota.handle_delete()
        self.quotas.delete.assert_called_once_with('some_project_id')

    def test_quota_with_invalid_gigabytes(self):
        fake_v = self.fv(2)
        self.volumes.list.return_value = [fake_v, fake_v]
        self.my_quota.physical_resource_name = mock.MagicMock(
            return_value='some_resource_id')
        self.my_quota.reparse()
        err = self.assertRaises(ValueError, self.my_quota.handle_create)
        self.assertEqual(
            self.err_msg % {'property': 'gigabytes', 'value': 5, 'total': 6},
            str(err))

    def test_quota_with_invalid_volumes(self):
        fake_v = self.fv(0)
        self.volumes.list.return_value = [fake_v, fake_v, fake_v, fake_v]
        self.my_quota.physical_resource_name = mock.MagicMock(
            return_value='some_resource_id')
        self.my_quota.reparse()
        err = self.assertRaises(ValueError, self.my_quota.handle_create)
        self.assertEqual(
            self.err_msg % {'property': 'volumes', 'value': 3, 'total': 4},
            str(err))

    def test_quota_with_invalid_snapshots(self):
        fake_v = self.fv(0)
        self.volume_snapshots.list.return_value = [fake_v, fake_v, fake_v,
                                                   fake_v]
        self.my_quota.physical_resource_name = mock.MagicMock(
            return_value='some_resource_id')
        self.my_quota.reparse()
        err = self.assertRaises(ValueError, self.my_quota.handle_create)
        self.assertEqual(
            self.err_msg % {'property': 'snapshots', 'value': 2, 'total': 4},
            str(err))

    def _test_quota_with_unlimited_value(self, prop_name):
        my_quota = self.stack['my_quota']
        props = self.stack.t.t['resources']['my_quota']['properties'].copy()
        props[prop_name] = -1
        my_quota.t = my_quota.t.freeze(properties=props)
        my_quota.reparse()
        my_quota.handle_create()
        kwargs = {'gigabytes': 5, 'snapshots': 2, 'volumes': 3}
        kwargs[prop_name] = -1
        self.quotas.update.assert_called_once_with('some_project_id', **kwargs)

    def test_quota_with_unlimited_gigabytes(self):
        self._test_quota_with_unlimited_value('gigabytes')

    def test_quota_with_unlimited_snapshots(self):
        self._test_quota_with_unlimited_value('snapshots')

    def test_quota_with_unlimited_volumes(self):
        self._test_quota_with_unlimited_value('volumes')



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""StackLock object."""

from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from heat.db import api as db_api
from heat.objects import base as heat_base


class StackLock(
        heat_base.HeatObject,
        base.VersionedObjectDictCompat,
        base.ComparableVersionedObject,
):
    fields = {
        'engine_id': fields.StringField(nullable=True),
        'stack_id': fields.StringField(),
        'created_at': fields.DateTimeField(read_only=True),
        'updated_at': fields.DateTimeField(nullable=True),
    }

    @classmethod
    def create(cls, context, stack_id, engine_id):
        return db_api.stack_lock_create(context, stack_id, engine_id)

    @classmethod
    def steal(cls, context, stack_id, old_engine_id, new_engine_id):
        return db_api.stack_lock_steal(context, stack_id,
                                       old_engine_id,
                                       new_engine_id)

    @classmethod
    def release(cls, context, stack_id, engine_id):
        return db_api.stack_lock_release(context, stack_id, engine_id)

    @classmethod
    def get_engine_id(cls, context, stack_id):
        return db_api.stack_lock_get_engine_id(context, stack_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support


class QoSPolicy(neutron.NeutronResource):
    """A resource for Neutron QoS Policy.

    This QoS policy can be associated with neutron resources,
    such as port and network, to provide QoS capabilities.

    The default policy usage of this resource is limited to
    administrators only.
    """

    required_service_extension = 'qos'

    entity = 'qos_policy'

    res_info_key = 'policy'

    support_status = support.SupportStatus(version='6.0.0')

    PROPERTIES = (
        NAME, DESCRIPTION, SHARED, TENANT_ID,
    ) = (
        'name', 'description', 'shared', 'tenant_id',
    )

    ATTRIBUTES = (
        RULES_ATTR,
    ) = (
        'rules',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('The name for the QoS policy.'),
            update_allowed=True
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('The description for the QoS policy.'),
            update_allowed=True
        ),
        SHARED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Whether this QoS policy should be shared to other tenants.'),
            default=False,
            update_allowed=True
        ),
        TENANT_ID: properties.Schema(
            properties.Schema.STRING,
            _('The owner tenant ID of this QoS policy.')
        ),
    }

    attributes_schema = {
        RULES_ATTR: attributes.Schema(
            _("A list of all rules for the QoS policy."),
            type=attributes.Schema.LIST
        )
    }

    def handle_create(self):
        props = self.prepare_properties(
            self.properties,
            self.physical_resource_name())

        policy = self.client().create_qos_policy({'policy': props})['policy']
        self.resource_id_set(policy['id'])

    def handle_delete(self):
        if self.resource_id is None:
            return

        with self.client_plugin().ignore_not_found:
            self.client().delete_qos_policy(self.resource_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.prepare_update_properties(prop_diff)
            self.client().update_qos_policy(
                self.resource_id,
                {'policy': prop_diff})


class QoSRule(neutron.NeutronResource):
    """A resource for Neutron QoS base rule."""

    required_service_extension = 'qos'

    support_status = support.SupportStatus(version='6.0.0')

    PROPERTIES = (
        POLICY, TENANT_ID,
    ) = (
        'policy', 'tenant_id',
    )

    properties_schema = {
        POLICY: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of the QoS policy.'),
            required=True,
            constraints=[constraints.CustomConstraint('neutron.qos_policy')]
        ),
        TENANT_ID: properties.Schema(
            properties.Schema.STRING,
            _('The owner tenant ID of this rule.')
        ),
    }

    def __init__(self, name, json_snippet, stack):
        super(QoSRule, self).__init__(name, json_snippet, stack)
        self._policy_id = None

    @property
    def policy_id(self):
        if not self._policy_id:
            self._policy_id = self.client_plugin().get_qos_policy_id(
                self.properties[self.POLICY])

        return self._policy_id


class QoSBandwidthLimitRule(QoSRule):
    """A resource for Neutron QoS bandwidth limit rule.

    This rule can be associated with QoS policy, and then the policy
    can be used by neutron port and network, to provide bandwidth limit
    QoS capabilities.

    The default policy usage of this resource is limited to
    administrators only.
    """

    entity = 'bandwidth_limit_rule'

    PROPERTIES = (
        MAX_BANDWIDTH, MAX_BURST_BANDWIDTH, DIRECTION
    ) = (
        'max_kbps', 'max_burst_kbps', 'direction'
    )

    properties_schema = {
        MAX_BANDWIDTH: properties.Schema(
            properties.Schema.INTEGER,
            _('Max bandwidth in kbps.'),
            required=True,
            update_allowed=True,
            constraints=[
                constraints.Range(min=0)
            ]
        ),
        MAX_BURST_BANDWIDTH: properties.Schema(
            properties.Schema.INTEGER,
            _('Max burst bandwidth in kbps.'),
            update_allowed=True,
            constraints=[
                constraints.Range(min=0)
            ],
            default=0
        ),
        DIRECTION: properties.Schema(
            properties.Schema.STRING,
            _('Traffic direction from the point of view of the port.'),
            update_allowed=True,
            constraints=[
                constraints.AllowedValues(['egress', 'ingress']),
            ],
            default='egress',
            support_status=support.SupportStatus(version='13.0.0')
        )
    }

    properties_schema.update(QoSRule.properties_schema)

    def handle_create(self):
        props = self.prepare_properties(self.properties,
                                        self.physical_resource_name())
        props.pop(self.POLICY)

        rule = self.client().create_bandwidth_limit_rule(
            self.policy_id,
            {'bandwidth_limit_rule': props})['bandwidth_limit_rule']

        self.resource_id_set(rule['id'])

    def handle_delete(self):
        if self.resource_id is None:
            return

        with self.client_plugin().ignore_not_found:
            self.client().delete_bandwidth_limit_rule(
                self.resource_id, self.policy_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.client().update_bandwidth_limit_rule(
                self.resource_id,
                self.policy_id,
                {'bandwidth_limit_rule': prop_diff})

    def _res_get_args(self):
        return [self.resource_id, self.policy_id]


class QoSDscpMarkingRule(QoSRule):
    """A resource for Neutron QoS DSCP marking rule.

    This rule can be associated with QoS policy, and then the policy
    can be used by neutron port and network, to provide DSCP marking
    QoS capabilities.

    The default policy usage of this resource is limited to
    administrators only.
    """

    support_status = support.SupportStatus(version='7.0.0')

    entity = 'dscp_marking_rule'

    PROPERTIES = (
        DSCP_MARK,
    ) = (
        'dscp_mark',
    )

    properties_schema = {
        DSCP_MARK: properties.Schema(
            properties.Schema.INTEGER,
            _('DSCP mark between 0 and 56, except 2-6, 42, 44, and 50-54.'),
            required=True,
            update_allowed=True,
            constraints=[
                constraints.AllowedValues([
                    0, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30, 32, 34,
                    36, 38, 40, 46, 48, 56]
                )
            ]
        )
    }

    properties_schema.update(QoSRule.properties_schema)

    def handle_create(self):
        props = self.prepare_properties(self.properties,
                                        self.physical_resource_name())
        props.pop(self.POLICY)

        rule = self.client().create_dscp_marking_rule(
            self.policy_id,
            {'dscp_marking_rule': props})['dscp_marking_rule']

        self.resource_id_set(rule['id'])

    def handle_delete(self):
        if self.resource_id is None:
            return

        with self.client_plugin().ignore_not_found:
            self.client().delete_dscp_marking_rule(
                self.resource_id, self.policy_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.client().update_dscp_marking_rule(
                self.resource_id,
                self.policy_id,
                {'dscp_marking_rule': prop_diff})

    def _res_get_args(self):
        return [self.resource_id, self.policy_id]


class QoSMinimumBandwidthRule(QoSRule):
    """A resource for guaranteeing bandwidth.

    This rule can be associated with a QoS policy, and then the policy
    can be used by a neutron port to provide guaranteed bandwidth QoS
    capabilities.

    Depending on drivers the guarantee may be enforced on two levels.
    First when a server is placed (scheduled) on physical infrastructure
    and/or second in the data plane of the physical hypervisor. For details
    please see Neutron documentation:

    https://docs.openstack.org/neutron/latest/admin/config-qos-min-bw.html

    The default policy usage of this resource is limited to
    administrators only.
    """

    entity = 'minimum_bandwidth_rule'

    required_service_extension = 'qos-bw-minimum-ingress'

    support_status = support.SupportStatus(
        status=support.SUPPORTED,
        version='14.0.0',
    )

    PROPERTIES = (
        MIN_BANDWIDTH, DIRECTION
    ) = (
        'min_kbps', 'direction'
    )

    properties_schema = {
        MIN_BANDWIDTH: properties.Schema(
            properties.Schema.INTEGER,
            _('Min bandwidth in kbps.'),
            required=True,
            update_allowed=True,
            constraints=[
                constraints.Range(min=0),
            ],
        ),
        DIRECTION: properties.Schema(
            properties.Schema.STRING,
            _('Traffic direction from the point of view of the port.'),
            update_allowed=True,
            constraints=[
                constraints.AllowedValues(['egress', 'ingress']),
            ],
            default='egress',
        ),
    }

    properties_schema.update(QoSRule.properties_schema)

    def handle_create(self):
        props = self.prepare_properties(self.properties,
                                        self.physical_resource_name())
        props.pop(self.POLICY)

        rule = self.client().create_minimum_bandwidth_rule(
            self.policy_id,
            {'minimum_bandwidth_rule': props})['minimum_bandwidth_rule']

        self.resource_id_set(rule['id'])

    def handle_delete(self):
        if self.resource_id is None:
            return

        with self.client_plugin().ignore_not_found:
            self.client().delete_minimum_bandwidth_rule(
                self.resource_id, self.policy_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.client().update_minimum_bandwidth_rule(
                self.resource_id,
                self.policy_id,
                {'minimum_bandwidth_rule': prop_diff})

    def _res_get_args(self):
        return [self.resource_id, self.policy_id]


class QoSMinimumPacketRateRule(QoSRule):
    """A resource for guaranteeing packet rate.

    This rule can be associated with a QoS policy, and then the policy
    can be used by a neutron port to provide guaranteed packet rate QoS
    capabilities.

    Depending on drivers the guarantee may be enforced on two levels.
    First when a server is placed (scheduled) on physical infrastructure
    and/or second in the data plane of the physical hypervisor. For details
    please see Neutron documentation:

    https://docs.openstack.org/neutron/latest/admin/config-qos-min-pps.html

    The default policy usage of this resource is limited to
    administrators only.
    """

    entity = 'minimum_packet_rate_rule'

    required_service_extension = 'qos-pps-minimum'

    support_status = support.SupportStatus(
        status=support.SUPPORTED,
        version='19.0.0',
    )

    PROPERTIES = (
        MIN_PACKET_RATE, DIRECTION
    ) = (
        'min_kpps', 'direction'
    )

    properties_schema = {
        MIN_PACKET_RATE: properties.Schema(
            properties.Schema.INTEGER,
            _('Min packet rate in kpps.'),
            required=True,
            update_allowed=True,
            constraints=[
                constraints.Range(min=0),
            ],
        ),
        DIRECTION: properties.Schema(
            properties.Schema.STRING,
            _('Traffic direction from the point of view of the port.'),
            update_allowed=True,
            constraints=[
                constraints.AllowedValues(['any', 'egress', 'ingress']),
            ],
            default='egress',
        ),
    }

    properties_schema.update(QoSRule.properties_schema)

    def handle_create(self):
        props = self.prepare_properties(self.properties,
                                        self.physical_resource_name())
        props.pop(self.POLICY)

        rule = self.client().create_minimum_packet_rate_rule(
            self.policy_id,
            {'minimum_packet_rate_rule': props})['minimum_packet_rate_rule']

        self.resource_id_set(rule['id'])

    def handle_delete(self):
        if self.resource_id is None:
            return

        with self.client_plugin().ignore_not_found:
            self.client().delete_minimum_packet_rate_rule(
                self.resource_id, self.policy_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.client().update_minimum_packet_rate_rule(
                self.resource_id,
                self.policy_id,
                {'minimum_packet_rate_rule': prop_diff})

    def _res_get_args(self):
        return [self.resource_id, self.policy_id]


def resource_mapping():
    return {
        'OS::Neutron::QoSPolicy': QoSPolicy,
        'OS::Neutron::QoSBandwidthLimitRule': QoSBandwidthLimitRule,
        'OS::Neutron::QoSDscpMarkingRule': QoSDscpMarkingRule,
        'OS::Neutron::QoSMinimumBandwidthRule': QoSMinimumBandwidthRule,
        'OS::Neutron::QoSMinimumPacketRateRule': QoSMinimumPacketRateRule,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import properties
from heat.engine import resource
from heat.engine.resources.aws.ec2 import vpc


class Subnet(resource.Resource):

    PROPERTIES = (
        AVAILABILITY_ZONE, CIDR_BLOCK, VPC_ID, TAGS,
    ) = (
        'AvailabilityZone', 'CidrBlock', 'VpcId', 'Tags',
    )

    _TAG_KEYS = (
        TAG_KEY, TAG_VALUE,
    ) = (
        'Key', 'Value',
    )

    ATTRIBUTES = (
        AVAILABILITY_ZONE,
    )

    properties_schema = {
        AVAILABILITY_ZONE: properties.Schema(
            properties.Schema.STRING,
            _('Availability zone in which you want the subnet.')
        ),
        CIDR_BLOCK: properties.Schema(
            properties.Schema.STRING,
            _('CIDR block to apply to subnet.'),
            required=True
        ),
        VPC_ID: properties.Schema(
            properties.Schema.STRING,
            _('Ref structure that contains the ID of the VPC on which you '
              'want to create the subnet.'),
            required=True
        ),
        TAGS: properties.Schema(
            properties.Schema.LIST,
            schema=properties.Schema(
                properties.Schema.MAP,
                _('List of tags to attach to this resource.'),
                schema={
                    TAG_KEY: properties.Schema(
                        properties.Schema.STRING,
                        required=True
                    ),
                    TAG_VALUE: properties.Schema(
                        properties.Schema.STRING,
                        required=True
                    ),
                },
                implemented=False,
            )
        ),
    }

    attributes_schema = {
        AVAILABILITY_ZONE: attributes.Schema(
            _('Availability Zone of the subnet.'),
            type=attributes.Schema.STRING
        ),
    }

    default_client_name = 'neutron'

    def handle_create(self):
        # TODO(sbaker) Verify that this CidrBlock is within the vpc CidrBlock
        network_id = self.properties.get(self.VPC_ID)

        props = {
            'network_id': network_id,
            'cidr': self.properties.get(self.CIDR_BLOCK),
            'name': self.physical_resource_name(),
            'ip_version': 4
        }
        subnet = self.client().create_subnet({'subnet': props})['subnet']
        self.resource_id_set(subnet['id'])

        router = vpc.VPC.router_for_vpc(self.client(), network_id)
        if router:
            self.client().add_interface_router(
                router['id'],
                {'subnet_id': subnet['id']})

    def handle_delete(self):
        if self.resource_id is None:
            return

        network_id = self.properties.get(self.VPC_ID)
        subnet_id = self.resource_id

        with self.client_plugin().ignore_not_found:
            router = vpc.VPC.router_for_vpc(self.client(), network_id)
            if router:
                self.client().remove_interface_router(
                    router['id'],
                    {'subnet_id': subnet_id})

        with self.client_plugin().ignore_not_found:
            self.client().delete_subnet(subnet_id)

    def _resolve_attribute(self, name):
        if name == self.AVAILABILITY_ZONE:
            return self.properties.get(self.AVAILABILITY_ZONE)


def resource_mapping():
    return {
        'AWS::EC2::Subnet': Subnet,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import function
from heat.engine import properties
from heat.engine import resource


class LaunchConfiguration(resource.Resource):

    PROPERTIES = (
        IMAGE_ID, INSTANCE_TYPE, KEY_NAME, USER_DATA, SECURITY_GROUPS,
        KERNEL_ID, RAM_DISK_ID, BLOCK_DEVICE_MAPPINGS, NOVA_SCHEDULER_HINTS,
        INSTANCE_ID,
    ) = (
        'ImageId', 'InstanceType', 'KeyName', 'UserData', 'SecurityGroups',
        'KernelId', 'RamDiskId', 'BlockDeviceMappings', 'NovaSchedulerHints',
        'InstanceId',
    )

    _NOVA_SCHEDULER_HINT_KEYS = (
        NOVA_SCHEDULER_HINT_KEY, NOVA_SCHEDULER_HINT_VALUE,
    ) = (
        'Key', 'Value',
    )

    _BLOCK_DEVICE_MAPPING_KEYS = (
        DEVICE_NAME, EBS, NO_DEVICE, VIRTUAL_NAME,
    ) = (
        'DeviceName', 'Ebs', 'NoDevice', 'VirtualName',
    )

    _EBS_KEYS = (
        DELETE_ON_TERMINATION, IOPS, SNAPSHOT_ID, VOLUME_SIZE,
        VOLUME_TYPE,
    ) = (
        'DeleteOnTermination', 'Iops', 'SnapshotId', 'VolumeSize',
        'VolumeType'
    )

    properties_schema = {
        IMAGE_ID: properties.Schema(
            properties.Schema.STRING,
            _('Glance image ID or name.'),
            constraints=[
                constraints.CustomConstraint('glance.image')
            ]
        ),
        INSTANCE_TYPE: properties.Schema(
            properties.Schema.STRING,
            _('Nova instance type (flavor).'),
            constraints=[
                constraints.CustomConstraint('nova.flavor')
            ]
        ),
        INSTANCE_ID: properties.Schema(
            properties.Schema.STRING,
            _('The ID of an existing instance you want to use to create '
              'the launch configuration. All properties are derived from '
              'the instance with the exception of BlockDeviceMapping.'),
            constraints=[
                constraints.CustomConstraint("nova.server")
            ]
        ),
        KEY_NAME: properties.Schema(
            properties.Schema.STRING,
            _('Optional Nova keypair name.'),
            constraints=[
                constraints.CustomConstraint("nova.keypair")
            ]
        ),
        USER_DATA: properties.Schema(
            properties.Schema.STRING,
            _('User data to pass to instance.')
        ),
        SECURITY_GROUPS: properties.Schema(
            properties.Schema.LIST,
            _('Security group names to assign.')
        ),
        KERNEL_ID: properties.Schema(
            properties.Schema.STRING,
            _('Not Implemented.'),
            implemented=False
        ),
        RAM_DISK_ID: properties.Schema(
            properties.Schema.STRING,
            _('Not Implemented.'),
            implemented=False
        ),
        BLOCK_DEVICE_MAPPINGS: properties.Schema(
            properties.Schema.LIST,
            _('Block device mappings to attach to instance.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    DEVICE_NAME: properties.Schema(
                        properties.Schema.STRING,
                        _('A device name where the volume will be '
                          'attached in the system at /dev/device_name.'
                          'e.g. vdb'),
                        required=True,
                    ),
                    EBS: properties.Schema(
                        properties.Schema.MAP,
                        _('The ebs volume to attach to the instance.'),
                        schema={
                            DELETE_ON_TERMINATION: properties.Schema(
                                properties.Schema.BOOLEAN,
                                _('Indicate whether the volume should be '
                                  'deleted when the instance is terminated.'),
                                default=True
                            ),
                            IOPS: properties.Schema(
                                properties.Schema.NUMBER,
                                _('The number of I/O operations per second '
                                  'that the volume supports.'),
                                implemented=False
                            ),
                            SNAPSHOT_ID: properties.Schema(
                                properties.Schema.STRING,
                                _('The ID of the snapshot to create '
                                  'a volume from.'),
                                constraints=[
                                    constraints.CustomConstraint(
                                        'cinder.snapshot')
                                ]
                            ),
                            VOLUME_SIZE: properties.Schema(
                                properties.Schema.STRING,
                                _('The size of the volume, in GB. Must be '
                                  'equal or greater than the size of the '
                                  'snapshot. It is safe to leave this blank '
                                  'and have the Compute service infer '
                                  'the size.'),
                            ),
                            VOLUME_TYPE: properties.Schema(
                                properties.Schema.STRING,
                                _('The volume type.'),
                                implemented=False
                            ),
                        },
                    ),
                    NO_DEVICE: properties.Schema(
                        properties.Schema.MAP,
                        _('The can be used to unmap a defined device.'),
                        implemented=False
                    ),
                    VIRTUAL_NAME: properties.Schema(
                        properties.Schema.STRING,
                        _('The name of the virtual device. The name must be '
                          'in the form ephemeralX where X is a number '
                          'starting from zero (0); for example, ephemeral0.'),
                        implemented=False
                    ),
                },
            ),
        ),
        NOVA_SCHEDULER_HINTS: properties.Schema(
            properties.Schema.LIST,
            _('Scheduler hints to pass to Nova (Heat extension).'),
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    NOVA_SCHEDULER_HINT_KEY: properties.Schema(
                        properties.Schema.STRING,
                        required=True
                    ),
                    NOVA_SCHEDULER_HINT_VALUE: properties.Schema(
                        properties.Schema.STRING,
                        required=True
                    ),
                },
            )
        ),
    }

    def rebuild_lc_properties(self, instance_id):
        server = self.client_plugin('nova').get_server(instance_id)
        instance_props = {
            self.IMAGE_ID: server.image['id'],
            self.INSTANCE_TYPE: server.flavor['id'],
            self.KEY_NAME: server.key_name,
            self.SECURITY_GROUPS: [sg['name']
                                   for sg in server.security_groups]
        }
        lc_props = function.resolve(self.properties.data)
        for key, value in instance_props.items():
            # the properties which are specified in launch configuration,
            # will override the attributes from the instance
            lc_props.setdefault(key, value)

        return lc_props

    def handle_create(self):
        instance_id = self.properties.get(self.INSTANCE_ID)
        if instance_id:
            lc_props = self.rebuild_lc_properties(instance_id)
            defn = self.t.freeze(properties=lc_props)
            self.properties = defn.properties(
                self.properties_schema, self.context)
            self._update_stored_properties()

    def needs_replace_with_tmpl_diff(self, tmpl_diff):
        return tmpl_diff.metadata_changed()

    def get_reference_id(self):
        return self.physical_resource_name_or_FnGetRefId()

    def validate(self):
        """Validate any of the provided params."""
        super(LaunchConfiguration, self).validate()
        # now we don't support without snapshot_id in bdm
        bdm = self.properties.get(self.BLOCK_DEVICE_MAPPINGS)
        if bdm:
            for mapping in bdm:
                ebs = mapping.get(self.EBS)
                if ebs:
                    snapshot_id = ebs.get(self.SNAPSHOT_ID)
                    if not snapshot_id:
                        msg = _("SnapshotId is missing, this is required "
                                "when specifying BlockDeviceMappings.")
                        raise exception.StackValidationFailed(message=msg)
                else:
                    msg = _("Ebs is missing, this is required "
                            "when specifying BlockDeviceMappings.")
                    raise exception.StackValidationFailed(message=msg)
        # validate the 'InstanceId', 'ImageId' and 'InstanceType',
        # if without 'InstanceId',  'ImageId' and 'InstanceType' are required
        instance_id = self.properties.get(self.INSTANCE_ID)
        if not instance_id:
            image_id = self.properties.get(self.IMAGE_ID)
            instance_type = self.properties.get(self.INSTANCE_TYPE)
            if not image_id or not instance_type:
                msg = _('If without InstanceId, '
                        'ImageId and InstanceType are required.')
                raise exception.StackValidationFailed(message=msg)


def resource_mapping():
    return {
        'AWS::AutoScaling::LaunchConfiguration': LaunchConfiguration,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutronclient.common import exceptions as qe
from neutronclient.neutron import v2_0 as neutronV20
from neutronclient.v2_0 import client as neutronclient

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import neutron
from heat.engine.resources.openstack.neutron import subnetpool
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.tests import common
from heat.tests.openstack.neutron import inline_templates
from heat.tests import utils


class NeutronSubnetPoolTest(common.HeatTestCase):

    def setUp(self):
        super(NeutronSubnetPoolTest, self).setUp()
        self.patchobject(neutron.NeutronClientPlugin, 'has_extension',
                         return_value=True)
        self.find_resource = self.patchobject(neutronV20,
                                              'find_resourceid_by_name_or_id',
                                              return_value='new_test')

    def create_subnetpool(self, status='COMPLETE', tags=None):
        self.t = template_format.parse(inline_templates.SPOOL_TEMPLATE)
        if tags:
            self.t['resources']['sub_pool']['properties']['tags'] = tags
        self.stack = utils.parse_stack(self.t)
        resource_defns = self.stack.t.resource_definitions(self.stack)
        rsrc = subnetpool.SubnetPool('sub_pool', resource_defns['sub_pool'],
                                     self.stack)
        if status == 'FAILED':
            self.patchobject(neutronclient.Client, 'create_subnetpool',
                             side_effect=qe.NeutronClientException(
                                 status_code=500))
            error = self.assertRaises(exception.ResourceFailure,
                                      scheduler.TaskRunner(rsrc.create))
            self.assertEqual(
                'NeutronClientException: resources.sub_pool: '
                'An unknown exception occurred.',
                str(error))
        else:
            self.patchobject(neutronclient.Client, 'create_subnetpool',
                             return_value={'subnetpool': {
                                 'id': 'fc68ea2c-b60b-4b4f-bd82-94ec81110766'
                             }})
            scheduler.TaskRunner(rsrc.create)()

        self.assertEqual((rsrc.CREATE, status), rsrc.state)
        if tags:
            self.set_tag_mock.assert_called_once_with('subnetpools',
                                                      rsrc.resource_id,
                                                      {'tags': tags})
        return rsrc

    def test_validate_prefixlen_min_gt_max(self):
        self.t = template_format.parse(inline_templates.SPOOL_TEMPLATE)
        props = self.t['resources']['sub_pool']['properties']
        props['min_prefixlen'] = 28
        props['max_prefixlen'] = 24
        self.stack = utils.parse_stack(self.t)
        rsrc = self.stack['sub_pool']
        errMessage = ('Illegal prefix bounds: max_prefixlen=24, '
                      'min_prefixlen=28.')
        error = self.assertRaises(exception.StackValidationFailed,
                                  rsrc.validate)
        self.assertEqual(errMessage, str(error))

    def test_validate_prefixlen_default_gt_max(self):
        self.t = template_format.parse(inline_templates.SPOOL_TEMPLATE)
        props = self.t['resources']['sub_pool']['properties']
        props['default_prefixlen'] = 28
        props['max_prefixlen'] = 24
        self.stack = utils.parse_stack(self.t)
        rsrc = self.stack['sub_pool']
        errMessage = ('Illegal prefix bounds: max_prefixlen=24, '
                      'default_prefixlen=28.')
        error = self.assertRaises(exception.StackValidationFailed,
                                  rsrc.validate)
        self.assertEqual(errMessage, str(error))

    def test_validate_prefixlen_min_gt_default(self):
        self.t = template_format.parse(inline_templates.SPOOL_TEMPLATE)
        props = self.t['resources']['sub_pool']['properties']
        props['min_prefixlen'] = 28
        props['default_prefixlen'] = 24
        self.stack = utils.parse_stack(self.t)
        rsrc = self.stack['sub_pool']
        errMessage = ('Illegal prefix bounds: min_prefixlen=28, '
                      'default_prefixlen=24.')
        error = self.assertRaises(exception.StackValidationFailed,
                                  rsrc.validate)
        self.assertEqual(errMessage, str(error))

    def test_validate_minimal(self):
        self.t = template_format.parse(inline_templates.SPOOL_MINIMAL_TEMPLATE)
        self.stack = utils.parse_stack(self.t)
        rsrc = self.stack['sub_pool']
        self.assertIsNone(rsrc.validate())

    def test_create_subnetpool(self):
        rsrc = self.create_subnetpool()
        ref_id = rsrc.FnGetRefId()
        self.assertEqual('fc68ea2c-b60b-4b4f-bd82-94ec81110766', ref_id)

    def test_create_subnetpool_with_tags(self):
        tags = ['for_test']
        self.set_tag_mock = self.patchobject(neutronclient.Client,
                                             'replace_tag')
        rsrc = self.create_subnetpool(tags=tags)
        ref_id = rsrc.FnGetRefId()
        self.assertEqual('fc68ea2c-b60b-4b4f-bd82-94ec81110766', ref_id)

    def test_create_subnetpool_failed(self):
        self.create_subnetpool('FAILED')

    def test_delete_subnetpool(self):
        self.patchobject(neutronclient.Client, 'delete_subnetpool')
        rsrc = self.create_subnetpool()
        ref_id = rsrc.FnGetRefId()
        self.assertEqual('fc68ea2c-b60b-4b4f-bd82-94ec81110766', ref_id)
        self.assertIsNone(scheduler.TaskRunner(rsrc.delete)())
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)

    def test_delete_subnetpool_not_found(self):
        self.patchobject(neutronclient.Client, 'delete_subnetpool',
                         side_effect=qe.NotFound(status_code=404))
        rsrc = self.create_subnetpool()
        ref_id = rsrc.FnGetRefId()
        self.assertEqual('fc68ea2c-b60b-4b4f-bd82-94ec81110766', ref_id)
        self.assertIsNone(scheduler.TaskRunner(rsrc.delete)())
        self.assertEqual((rsrc.DELETE, rsrc.COMPLETE), rsrc.state)

    def test_delete_subnetpool_resource_id_none(self):
        delete_pool = self.patchobject(neutronclient.Client,
                                       'delete_subnetpool')
        rsrc = self.create_subnetpool()
        rsrc.resource_id = None
        self.assertIsNone(scheduler.TaskRunner(rsrc.delete)())
        delete_pool.assert_not_called()

    def test_update_subnetpool(self):
        update_subnetpool = self.patchobject(neutronclient.Client,
                                             'update_subnetpool')
        self.set_tag_mock = self.patchobject(neutronclient.Client,
                                             'replace_tag')
        old_tags = ['old_tag']
        rsrc = self.create_subnetpool(tags=old_tags)
        self.patchobject(rsrc, 'physical_resource_name',
                         return_value='the_new_sp')
        ref_id = rsrc.FnGetRefId()
        self.assertEqual('fc68ea2c-b60b-4b4f-bd82-94ec81110766', ref_id)
        new_tags = ['new_tag']
        props = {
            'name': 'the_new_sp',
            'prefixes': [
                '10.1.0.0/16',
                '10.2.0.0/16'],
            'address_scope': 'new_test',
            'default_quota': '16',
            'default_prefixlen': '24',
            'min_prefixlen': '24',
            'max_prefixlen': '28',
            'is_default': False,
            'tags': new_tags
        }
        update_dict = props.copy()
        update_dict['name'] = 'the_new_sp'
        update_dict['address_scope_id'] = update_dict.pop('address_scope')
        update_snippet = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(),
                                                      props)
        # with name
        self.assertIsNone(rsrc.handle_update(update_snippet, {}, props))
        self.set_tag_mock.assert_called_with('subnetpools',
                                             rsrc.resource_id,
                                             {'tags': new_tags})

        # without name
        props['name'] = None
        self.assertIsNone(rsrc.handle_update(update_snippet, {}, props))

        self.assertEqual(2, update_subnetpool.call_count)
        update_dict.pop('tags')
        update_subnetpool.assert_called_with(
            'fc68ea2c-b60b-4b4f-bd82-94ec81110766',
            {'subnetpool': update_dict})

    def test_update_subnetpool_no_prop_diff(self):
        update_subnetpool = self.patchobject(neutronclient.Client,
                                             'update_subnetpool')
        rsrc = self.create_subnetpool()
        ref_id = rsrc.FnGetRefId()
        self.assertEqual('fc68ea2c-b60b-4b4f-bd82-94ec81110766', ref_id)
        update_snippet = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(),
                                                      rsrc.t._properties)
        self.assertIsNone(rsrc.handle_update(update_snippet, {}, {}))
        update_subnetpool.assert_not_called()

    def test_update_subnetpool_validate_prefixes(self):
        update_subnetpool = self.patchobject(neutronclient.Client,
                                             'update_subnetpool')
        rsrc = self.create_subnetpool()
        ref_id = rsrc.FnGetRefId()
        self.assertEqual('fc68ea2c-b60b-4b4f-bd82-94ec81110766', ref_id)
        prefix_old = rsrc.properties['prefixes']
        props = {
            'name': 'the_new_sp',
            'prefixes': ['10.5.0.0/16']
        }
        prefix_new = props['prefixes']
        update_snippet = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(),
                                                      props)
        errMessage = ('Property prefixes updated value %(value1)s '
                      'should be superset of existing value %(value2)s.'
                      % dict(value1=sorted(prefix_new),
                             value2=sorted(prefix_old)))

        error = self.assertRaises(exception.StackValidationFailed,
                                  rsrc.handle_update,
                                  update_snippet, {}, props)

        self.assertEqual(errMessage, str(error))
        update_subnetpool.assert_not_called()

        props = {
            'name': 'the_new_sp',
            'prefixes': ['10.0.0.0/8',
                         '10.6.0.0/16'],
        }

        update_snippet = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(),
                                                      props)
        self.assertIsNone(rsrc.handle_update(update_snippet, {}, props))
        update_subnetpool.assert_called_once_with(
            'fc68ea2c-b60b-4b4f-bd82-94ec81110766',
            {'subnetpool': props})

    def test_update_subnetpool_update_address_scope(self):
        update_subnetpool = self.patchobject(neutronclient.Client,
                                             'update_subnetpool')
        rsrc = self.create_subnetpool()
        ref_id = rsrc.FnGetRefId()
        self.assertEqual('fc68ea2c-b60b-4b4f-bd82-94ec81110766', ref_id)
        props = {
            'name': 'the_new_sp',
            'address_scope': 'new_test',
            'prefixes': ['10.0.0.0/8',
                         '10.6.0.0/16'],
        }
        update_dict = {
            'name': 'the_new_sp',
            'address_scope_id': 'new_test',
            'prefixes': ['10.0.0.0/8',
                         '10.6.0.0/16'],
        }
        update_snippet = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(),
                                                      props)
        self.assertIsNone(rsrc.handle_update(update_snippet, {}, props))
        self.assertEqual(3, self.find_resource.call_count)
        update_subnetpool.assert_called_once_with(
            'fc68ea2c-b60b-4b4f-bd82-94ec81110766',
            {'subnetpool': update_dict})

    def test_update_subnetpool_remove_address_scope(self):
        update_subnetpool = self.patchobject(neutronclient.Client,
                                             'update_subnetpool')
        rsrc = self.create_subnetpool()
        ref_id = rsrc.FnGetRefId()
        self.assertEqual('fc68ea2c-b60b-4b4f-bd82-94ec81110766', ref_id)
        props = {
            'name': 'the_new_sp',
            'prefixes': ['10.0.0.0/8',
                         '10.6.0.0/16'],
        }
        props_diff = {'address_scope': None}
        update_snippet = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(),
                                                      props)
        self.assertIsNone(rsrc.handle_update(update_snippet, {}, props_diff))
        self.assertEqual(2, self.find_resource.call_count)
        update_subnetpool.assert_called_once_with(
            'fc68ea2c-b60b-4b4f-bd82-94ec81110766',
            {'subnetpool': props_diff})



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import net
from heat.engine import support


class ProviderNet(net.Net):
    """A resource for managing Neutron provider networks.

    Provider networks specify details of physical realisation of the existing
    network.

    The default policy usage of this resource is limited to
    administrators only.
    """

    required_service_extension = 'provider'

    support_status = support.SupportStatus(version='2014.1')

    PROPERTIES = (
        NAME, PROVIDER_NETWORK_TYPE, PROVIDER_PHYSICAL_NETWORK,
        PROVIDER_SEGMENTATION_ID, ADMIN_STATE_UP, SHARED,
        PORT_SECURITY_ENABLED, ROUTER_EXTERNAL, DNS_DOMAIN,
        AVAILABILITY_ZONE_HINTS, TAGS, TENANT_ID,
    ) = (
        'name', 'network_type', 'physical_network',
        'segmentation_id', 'admin_state_up', 'shared',
        'port_security_enabled', 'router_external', 'dns_domain',
        'availability_zone_hints', 'tags', 'tenant_id',
    )

    ATTRIBUTES = (
        STATUS, SUBNETS, SEGMENTS,
    ) = (
        'status', 'subnets', 'segments',
    )

    NETWORK_TYPES = (
        LOCAL, VLAN, VXLAN, GRE, GENEVE, FLAT
    ) = (
        'local', 'vlan', 'vxlan', 'gre', 'geneve', 'flat'
    )

    properties_schema = {
        NAME: net.Net.properties_schema[NAME],
        PROVIDER_NETWORK_TYPE: properties.Schema(
            properties.Schema.STRING,
            _('A string specifying the provider network type for the '
              'network.'),
            update_allowed=True,
            required=True,
            constraints=[
                constraints.AllowedValues(NETWORK_TYPES),
            ]
        ),
        PROVIDER_PHYSICAL_NETWORK: properties.Schema(
            properties.Schema.STRING,
            _('A string specifying physical network mapping for the '
              'network.'),
            update_allowed=True
        ),
        PROVIDER_SEGMENTATION_ID: properties.Schema(
            properties.Schema.STRING,
            _('A string specifying the segmentation id for the '
              'network.'),
            update_allowed=True
        ),
        ADMIN_STATE_UP: net.Net.properties_schema[ADMIN_STATE_UP],
        SHARED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Whether this network should be shared across all tenants.'),
            default=True,
            update_allowed=True
        ),
        PORT_SECURITY_ENABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Flag to enable/disable port security on the network. It '
              'provides the default value for the attribute of the ports '
              'created on this network.'),
            update_allowed=True,
            support_status=support.SupportStatus(version='8.0.0')
        ),
        ROUTER_EXTERNAL: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Whether the network contains an external router.'),
            default=False,
            update_allowed=True,
            support_status=support.SupportStatus(version='6.0.0')
        ),
        TAGS: properties.Schema(
            properties.Schema.LIST,
            _('The tags to be added to the provider network.'),
            schema=properties.Schema(properties.Schema.STRING),
            update_allowed=True,
            support_status=support.SupportStatus(version='12.0.0')
        ),
        DNS_DOMAIN: properties.Schema(
            properties.Schema.STRING,
            _('DNS domain associated with this network.'),
            constraints=[
                constraints.CustomConstraint('dns_domain')
            ],
            update_allowed=True,
            support_status=support.SupportStatus(version='15.0.0')
        ),
        AVAILABILITY_ZONE_HINTS: properties.Schema(
            properties.Schema.LIST,
            _('Availability zone candidates for the network. It requires the '
              'availability_zone extension to be available.'),
            update_allowed=True,
            support_status=support.SupportStatus(version='19.0.0')
        ),
        TENANT_ID: properties.Schema(
            properties.Schema.STRING,
            _('The ID of the tenant which will own the provider network. Only '
              'administrative users can set the tenant identifier; this '
              'cannot be changed using authorization policies.'),
            support_status=support.SupportStatus(version='24.0.0')
        ),
    }

    attributes_schema = {
        STATUS: attributes.Schema(
            _("The status of the network."),
            type=attributes.Schema.STRING
        ),
        SUBNETS: attributes.Schema(
            _("Subnets of this network."),
            type=attributes.Schema.LIST
        ),
        SEGMENTS: attributes.Schema(
            _("The segments of this network."),
            type=attributes.Schema.LIST,
            support_status=support.SupportStatus(version='16.0.0'),
        ),
    }

    def validate(self):
        """Resource's validation.

        Validates to ensure that segmentation_id is not there for flat
        network type.
        """
        super(ProviderNet, self).validate()

        if (self.properties[self.PROVIDER_SEGMENTATION_ID] and
                self.properties[self.PROVIDER_NETWORK_TYPE] in (
                    self.LOCAL, self.FLAT)):
            msg = _(
                'segmentation_id not allowed for network types: local, flat')
            raise exception.StackValidationFailed(message=msg)

        if (self.properties[self.PROVIDER_PHYSICAL_NETWORK] and
                self.properties[self.PROVIDER_NETWORK_TYPE] in (
                    self.VXLAN, self.GRE, self.GENEVE, self.LOCAL)):
            msg = _(
                'physical_network not allowed for network types: '
                'vxlan, gre, geneve, flat')
            raise exception.StackValidationFailed(message=msg)

    @staticmethod
    def add_provider_extension(props, key):
        props['provider:' + key] = props.pop(key)

    @staticmethod
    def prepare_provider_properties(props):
        if ProviderNet.PROVIDER_NETWORK_TYPE in props:
            ProviderNet.add_provider_extension(
                props,
                ProviderNet.PROVIDER_NETWORK_TYPE)
        if ProviderNet.PROVIDER_PHYSICAL_NETWORK in props:
            ProviderNet.add_provider_extension(
                props,
                ProviderNet.PROVIDER_PHYSICAL_NETWORK)
        if ProviderNet.PROVIDER_SEGMENTATION_ID in props:
            ProviderNet.add_provider_extension(
                props,
                ProviderNet.PROVIDER_SEGMENTATION_ID)

        if ProviderNet.ROUTER_EXTERNAL in props:
            props['router:external'] = props.pop(ProviderNet.ROUTER_EXTERNAL)

    def handle_create(self):
        """Creates the resource with provided properties.

        Adds 'provider:' extension to the required properties during create.
        """
        props = self.prepare_properties(
            self.properties,
            self.physical_resource_name())

        ProviderNet.prepare_provider_properties(props)
        tags = props.pop(self.TAGS, [])

        prov_net = self.client().create_network({'network': props})['network']
        self.resource_id_set(prov_net['id'])

        if tags:
            self.set_tags(tags)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        """Updates the resource with provided properties.

        Adds 'provider:' extension to the required properties during update.
        """
        if prop_diff:
            ProviderNet.prepare_provider_properties(prop_diff)
            self.prepare_update_properties(prop_diff)
            if self.TAGS in prop_diff:
                self.set_tags(prop_diff.pop(self.TAGS))
        if prop_diff:
            self.client().update_network(self.resource_id,
                                         {'network': prop_diff})

    def parse_live_resource_data(self, resource_properties, resource_data):
        # this resource should not have super in case of we don't need to
        # parse Net resource properties.
        result = {}
        provider_keys = [self.PROVIDER_NETWORK_TYPE,
                         self.PROVIDER_PHYSICAL_NETWORK,
                         self.PROVIDER_SEGMENTATION_ID]
        for key in provider_keys:
            result[key] = resource_data.get('provider:%s' % key)
        result[self.ROUTER_EXTERNAL] = resource_data.get('router:external')
        provider_keys.append(self.ROUTER_EXTERNAL)

        provider_keys.append(self.SHARED)
        for key in set(self.PROPERTIES) - set(provider_keys):
            if key in resource_data:
                result[key] = resource_data.get(key)
        return result


def resource_mapping():
    return {
        'OS::Neutron::ProviderNet': ProviderNet,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
from oslo_messaging.rpc import dispatcher

from heat.common import exception
from heat.engine import service
from heat.engine import stack as parser
from heat.objects import stack as stack_object
from heat.tests import common
from heat.tests import utils


class StackServiceAdoptTest(common.HeatTestCase):

    def setUp(self):
        super(StackServiceAdoptTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.man = service.EngineService('a-host', 'a-topic')
        self.man.thread_group_mgr = service.ThreadGroupManager()

    def _get_adopt_data_and_template(self, environment=None):
        template = {
            "heat_template_version": "2013-05-23",
            "parameters": {"app_dbx": {"type": "string"}},
            "resources": {"res1": {"type": "GenericResourceType"}}}

        adopt_data = {
            "status": "COMPLETE",
            "name": "rtrove1",
            "environment": environment,
            "template": template,
            "action": "CREATE",
            "id": "8532f0d3-ea84-444e-b2bb-2543bb1496a4",
            "resources": {"res1": {
                    "status": "COMPLETE",
                    "name": "database_password",
                    "resource_id": "yBpuUROjfGQ2gKOD",
                    "action": "CREATE",
                    "type": "GenericResourceType",
                    "metadata": {}}}}
        return template, adopt_data

    def _do_adopt(self, stack_name, template, input_params, adopt_data):
        result = self.man.create_stack(self.ctx, stack_name,
                                       template, input_params, None,
                                       {'adopt_stack_data': str(adopt_data)})
        self.man.thread_group_mgr.stop(result['stack_id'], graceful=True)
        return result

    def test_stack_adopt_with_params(self):
        cfg.CONF.set_override('enable_stack_adopt', True)
        cfg.CONF.set_override('convergence_engine', False)
        env = {'parameters': {"app_dbx": "test"}}
        template, adopt_data = self._get_adopt_data_and_template(env)
        result = self._do_adopt("test_adopt_with_params", template, {},
                                adopt_data)

        stack = stack_object.Stack.get_by_id(self.ctx, result['stack_id'])
        self.assertEqual(template, stack.raw_template.template)
        self.assertEqual(env['parameters'],
                         stack.raw_template.environment['parameters'])

    @mock.patch.object(parser.Stack, '_converge_create_or_update')
    @mock.patch.object(parser.Stack, '_send_notification_and_add_event')
    def test_convergence_stack_adopt_with_params(self,
                                                 mock_converge,
                                                 mock_send_notif):
        cfg.CONF.set_override('enable_stack_adopt', True)
        cfg.CONF.set_override('convergence_engine', True)
        env = {'parameters': {"app_dbx": "test"}}
        template, adopt_data = self._get_adopt_data_and_template(env)
        result = self._do_adopt("test_adopt_with_params", template, {},
                                adopt_data)

        stack = stack_object.Stack.get_by_id(self.ctx, result['stack_id'])
        self.assertEqual(template, stack.raw_template.template)
        self.assertEqual(env['parameters'],
                         stack.raw_template.environment['parameters'])
        self.assertTrue(mock_converge.called)

    def test_stack_adopt_saves_input_params(self):
        cfg.CONF.set_override('enable_stack_adopt', True)
        cfg.CONF.set_override('convergence_engine', False)
        env = {'parameters': {"app_dbx": "foo"}}
        input_params = {
            "parameters": {"app_dbx": "bar"}
        }
        template, adopt_data = self._get_adopt_data_and_template(env)
        result = self._do_adopt("test_adopt_saves_inputs", template,
                                input_params, adopt_data)

        stack = stack_object.Stack.get_by_id(self.ctx, result['stack_id'])
        self.assertEqual(template, stack.raw_template.template)
        self.assertEqual(input_params['parameters'],
                         stack.raw_template.environment['parameters'])

    @mock.patch.object(parser.Stack, '_converge_create_or_update')
    @mock.patch.object(parser.Stack, '_send_notification_and_add_event')
    def test_convergence_stack_adopt_saves_input_params(
            self, mock_converge, mock_send_notif):
        cfg.CONF.set_override('enable_stack_adopt', True)
        cfg.CONF.set_override('convergence_engine', True)
        env = {'parameters': {"app_dbx": "foo"}}
        input_params = {
            "parameters": {"app_dbx": "bar"}
        }
        template, adopt_data = self._get_adopt_data_and_template(env)
        result = self._do_adopt("test_adopt_saves_inputs", template,
                                input_params, adopt_data)

        stack = stack_object.Stack.get_by_id(self.ctx, result['stack_id'])
        self.assertEqual(template, stack.raw_template.template)
        self.assertEqual(input_params['parameters'],
                         stack.raw_template.environment['parameters'])
        self.assertTrue(mock_converge.called)

    def test_stack_adopt_stack_state(self):
        cfg.CONF.set_override('enable_stack_adopt', True)
        cfg.CONF.set_override('convergence_engine', False)
        env = {'parameters': {"app_dbx": "test"}}
        template, adopt_data = self._get_adopt_data_and_template(env)
        result = self._do_adopt("test_adopt_stack_state", template, {},
                                adopt_data)

        stack = stack_object.Stack.get_by_id(self.ctx, result['stack_id'])
        self.assertEqual((parser.Stack.ADOPT, parser.Stack.COMPLETE),
                         (stack.action, stack.status))

    @mock.patch.object(parser.Stack, '_converge_create_or_update')
    @mock.patch.object(parser.Stack, '_send_notification_and_add_event')
    def test_convergence_stack_adopt_stack_state(self, mock_converge,
                                                 mock_send_notif):
        cfg.CONF.set_override('enable_stack_adopt', True)
        cfg.CONF.set_override('convergence_engine', True)
        env = {'parameters': {"app_dbx": "test"}}
        template, adopt_data = self._get_adopt_data_and_template(env)
        result = self._do_adopt("test_adopt_stack_state", template, {},
                                adopt_data)

        stack = stack_object.Stack.get_by_id(self.ctx, result['stack_id'])
        self.assertEqual((parser.Stack.ADOPT, parser.Stack.IN_PROGRESS),
                         (stack.action, stack.status))
        self.assertTrue(mock_converge.called)

    def test_stack_adopt_disabled(self):
        # to test disable stack adopt
        cfg.CONF.set_override('enable_stack_adopt', False)
        env = {'parameters': {"app_dbx": "test"}}
        template, adopt_data = self._get_adopt_data_and_template(env)
        ex = self.assertRaises(
            dispatcher.ExpectedException,
            self.man.create_stack,
            self.ctx, "test_adopt_stack_disabled",
            template, {}, None,
            {'adopt_stack_data': str(adopt_data)})
        self.assertEqual(exception.NotSupported, ex.exc_info[0])
        self.assertIn('Stack Adopt', str(ex.exc_info[1]))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.tests import common


class TestException(common.HeatTestCase):

    class SampleExceptionNoErorCode(exception.HeatException):
        msg_fmt = 'Test exception'

    class SampleException(SampleExceptionNoErorCode):
        error_code = 100

    def test_heat_exception_no_error_code(self):
        ex = TestException.SampleExceptionNoErorCode()
        self.assertEqual('Test exception',
                         ex.message)

    def test_heat_exception_with_error_code(self):
        ex = TestException.SampleException()
        self.assertEqual('HEAT-E100 Test exception',
                         ex.message)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from heat_integrationtests.functional import functional_base

test_template = {
    'heat_template_version': '2013-05-23',
    'description': 'Test template to create one instance.',
    'resources': {
        'bar': {
            'type': 'OS::Heat::TestResource',
            'properties': {
                'value': '1234',
                'update_replace': False,
            }
        }
    }
}

env_both_restrict = {'resource_registry': {
    'resources': {
        'bar': {'restricted_actions': ['update', 'replace']}
    }
}
}

env_replace_restrict = {'resource_registry': {
    'resources': {
        '*ar': {'restricted_actions': 'replace'}
    }
}
}

reason_update_restrict = 'update is restricted for resource.'
reason_replace_restrict = 'replace is restricted for resource.'


class UpdateRestrictedStackTest(functional_base.FunctionalTestsBase):

    def _check_for_restriction_reason(self, events,
                                      reason, num_expected=1):
        matched = [e for e in events
                   if e.resource_status_reason == reason]
        return len(matched) == num_expected

    def test_update(self):
        stack_identifier = self.stack_create(template=test_template)

        update_template = copy.deepcopy(test_template)
        props = update_template['resources']['bar']['properties']
        props['value'] = '4567'

        # check update fails - with 'both' restricted
        self.update_stack(stack_identifier, update_template,
                          env_both_restrict,
                          expected_status='UPDATE_FAILED')

        self.assertTrue(self.verify_resource_status(stack_identifier, 'bar',
                                                    'CREATE_COMPLETE'))
        resource_events = self.client.events.list(stack_identifier, 'bar')
        self.assertTrue(
            self._check_for_restriction_reason(resource_events,
                                               reason_update_restrict))

        # check update succeeds - with only 'replace' restricted
        self.update_stack(stack_identifier, update_template,
                          env_replace_restrict,
                          expected_status='UPDATE_COMPLETE')

        self.assertTrue(self.verify_resource_status(stack_identifier, 'bar',
                                                    'UPDATE_COMPLETE'))
        resource_events = self.client.events.list(stack_identifier, 'bar')
        self.assertFalse(
            self._check_for_restriction_reason(resource_events,
                                               reason_update_restrict, 2))
        self.assertTrue(
            self._check_for_restriction_reason(resource_events,
                                               reason_replace_restrict, 0))

    def test_replace(self):
        stack_identifier = self.stack_create(template=test_template)

        update_template = copy.deepcopy(test_template)
        props = update_template['resources']['bar']['properties']
        props['value'] = '4567'
        props['update_replace'] = True

        # check replace fails - with 'both' restricted
        self.update_stack(stack_identifier, update_template,
                          env_both_restrict,
                          expected_status='UPDATE_FAILED')

        self.assertTrue(self.verify_resource_status(stack_identifier, 'bar',
                                                    'CREATE_COMPLETE'))
        resource_events = self.client.events.list(stack_identifier, 'bar')
        self.assertTrue(
            self._check_for_restriction_reason(resource_events,
                                               reason_replace_restrict))

        # check replace fails - with only 'replace' restricted
        self.update_stack(stack_identifier, update_template,
                          env_replace_restrict,
                          expected_status='UPDATE_FAILED')

        self.assertTrue(self.verify_resource_status(stack_identifier, 'bar',
                                                    'CREATE_COMPLETE'))
        resource_events = self.client.events.list(stack_identifier, 'bar')
        self.assertTrue(
            self._check_for_restriction_reason(resource_events,
                                               reason_replace_restrict, 2))
        self.assertTrue(
            self._check_for_restriction_reason(resource_events,
                                               reason_update_restrict, 0))

    def test_update_type_changed(self):
        stack_identifier = self.stack_create(template=test_template)

        update_template = copy.deepcopy(test_template)
        rsrc = update_template['resources']['bar']
        rsrc['type'] = 'OS::Heat::None'

        # check replace fails - with 'both' restricted
        self.update_stack(stack_identifier, update_template,
                          env_both_restrict,
                          expected_status='UPDATE_FAILED')

        self.assertTrue(self.verify_resource_status(stack_identifier, 'bar',
                                                    'CREATE_COMPLETE'))
        resource_events = self.client.events.list(stack_identifier, 'bar')
        self.assertTrue(
            self._check_for_restriction_reason(resource_events,
                                               reason_replace_restrict))

        # check replace fails - with only 'replace' restricted
        self.update_stack(stack_identifier, update_template,
                          env_replace_restrict,
                          expected_status='UPDATE_FAILED')

        self.assertTrue(self.verify_resource_status(stack_identifier, 'bar',
                                                    'CREATE_COMPLETE'))
        resource_events = self.client.events.list(stack_identifier, 'bar')
        self.assertTrue(
            self._check_for_restriction_reason(resource_events,
                                               reason_replace_restrict, 2))
        self.assertTrue(
            self._check_for_restriction_reason(resource_events,
                                               reason_update_restrict, 0))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.engine import stack
from heat.engine import template
from heat.tests import common
from heat.tests import utils

flavor_template = {
    'heat_template_version': '2013-05-23',
    'resources': {
        'my_flavor': {
            'type': 'OS::Nova::Flavor',
            'properties': {
                'ram': 1024,
                'vcpus': 2,
                'disk': 20,
                'swap': 2,
                'rxtx_factor': 1.0,
                'ephemeral': 0,
                'extra_specs': {"foo": "bar"},
                'tenants': []
            }
        }
    }
}


class NovaFlavorTest(common.HeatTestCase):
    def setUp(self):
        super(NovaFlavorTest, self).setUp()
        self.ctx = utils.dummy_context()

    def create_flavor(self, with_name_id=False, is_public=True):
        if with_name_id:
            props = flavor_template['resources']['my_flavor']['properties']
            props['name'] = 'test_flavor'
            props['flavorid'] = '1234'
        if not is_public:
            props = flavor_template['resources']['my_flavor']['properties']
            props['is_public'] = False
            props['tenants'] = ["foo", "bar"]
        self.stack = stack.Stack(
            self.ctx, 'nova_flavor_test_stack',
            template.Template(flavor_template)
        )
        self.my_flavor = self.stack['my_flavor']
        nova = mock.MagicMock()
        self.novaclient = mock.MagicMock()
        self.my_flavor.client = nova
        nova.return_value = self.novaclient
        self.flavors = self.novaclient.flavors

    def test_flavor_handle_create_no_id_name(self):
        self.create_flavor()
        kwargs = {
            'vcpus': 2, 'disk': 20, 'swap': 2,
            'flavorid': 'auto', 'is_public': True,
            'rxtx_factor': 1.0, 'ram': 1024,
            'ephemeral': 0, 'name': 'm1.xxx'
        }
        self.patchobject(self.my_flavor, 'physical_resource_name',
                         return_value='m1.xxx')
        value = mock.MagicMock()
        flavor_id = '927202df-1afb-497f-8368-9c2d2f26e5db'
        value.id = flavor_id
        value.is_public = True
        value.get_keys.return_value = {'k': 'v'}
        self.flavors.create.return_value = value
        self.flavors.get.return_value = value
        self.my_flavor.handle_create()
        self.flavors.create.assert_called_once_with(**kwargs)
        value.set_keys.assert_called_once_with({"foo": "bar"})
        self.assertEqual(flavor_id, self.my_flavor.resource_id)
        self.assertTrue(self.my_flavor.FnGetAtt('is_public'))
        self.assertEqual({'k': 'v'}, self.my_flavor.FnGetAtt('extra_specs'))

    def test_flavor_handle_create_with_id_name(self):
        self.create_flavor(with_name_id=True)
        kwargs = {
            'vcpus': 2, 'disk': 20, 'swap': 2,
            'flavorid': '1234', 'is_public': True,
            'rxtx_factor': 1.0, 'ram': 1024,
            'ephemeral': 0, 'name': 'test_flavor'
        }
        self.patchobject(self.my_flavor, 'physical_resource_name',
                         return_value='m1.xxx')
        value = mock.MagicMock()
        flavor_id = '927202df-1afb-497f-8368-9c2d2f26e5db'
        value.id = flavor_id
        value.is_public = True
        self.flavors.create.return_value = value
        self.flavors.get.return_value = value
        self.my_flavor.handle_create()
        self.flavors.create.assert_called_once_with(**kwargs)
        value.set_keys.assert_called_once_with({"foo": "bar"})
        self.assertEqual(flavor_id, self.my_flavor.resource_id)
        self.assertTrue(self.my_flavor.FnGetAtt('is_public'))

    def test_private_flavor_handle_create(self):
        self.create_flavor(is_public=False)
        value = mock.MagicMock()
        flavor_id = '927202df-1afb-497f-8368-9c2d2f26e5db'
        value.id = flavor_id
        value.is_public = False
        self.flavors.create.return_value = value
        self.flavors.get.return_value = value
        self.my_flavor.handle_create()
        value.set_keys.assert_called_once_with({"foo": "bar"})
        self.assertEqual(flavor_id, self.my_flavor.resource_id)
        self.assertFalse(self.my_flavor.FnGetAtt('is_public'))
        client_test = self.my_flavor.client().flavor_access.add_tenant_access
        test_tenants = [mock.call(value, 'foo'), mock.call(value, 'bar')]
        self.assertEqual(test_tenants, client_test.call_args_list)

    def test_flavor_handle_update_keys(self):
        self.create_flavor()
        value = mock.MagicMock()
        self.flavors.get.return_value = value
        value.get_keys.return_value = {}

        new_keys = {"new_foo": "new_bar"}
        prop_diff = {'extra_specs': new_keys}
        self.my_flavor.handle_update(json_snippet=None,
                                     tmpl_diff=None, prop_diff=prop_diff)
        value.unset_keys.assert_called_once_with({})
        value.set_keys.assert_called_once_with(new_keys)

    def test_flavor_handle_update_add_tenants(self):
        self.create_flavor(is_public=False)
        value = mock.MagicMock()

        new_tenants = ["new_foo", "new_bar"]
        prop_diff = {'tenants': new_tenants}
        self.flavors.get.return_value = value

        self.my_flavor.handle_update(json_snippet=None,
                                     tmpl_diff=None, prop_diff=prop_diff)
        test_tenants_add = [mock.call(value, 'new_foo'),
                            mock.call(value, 'new_bar')]
        test_add = self.my_flavor.client().flavor_access.add_tenant_access
        self.assertCountEqual(test_tenants_add,
                              test_add.call_args_list)

    def test_flavor_handle_update_remove_tenants(self):
        self.create_flavor(is_public=False)
        value = mock.MagicMock()

        new_tenants = []
        prop_diff = {'tenants': new_tenants}
        self.flavors.get.return_value = value

        itemFoo = mock.MagicMock()
        itemFoo.tenant_id = 'foo'
        itemBar = mock.MagicMock()
        itemBar.tenant_id = 'bar'

        self.my_flavor.client().flavor_access.list.return_value = [itemFoo,
                                                                   itemBar]
        self.my_flavor.handle_update(json_snippet=None,
                                     tmpl_diff=None, prop_diff=prop_diff)
        test_tenants_remove = [mock.call(value, 'foo'),
                               mock.call(value, 'bar')]
        test_rem = self.my_flavor.client().flavor_access.remove_tenant_access
        self.assertCountEqual(test_tenants_remove,
                              test_rem.call_args_list)

    def test_flavor_show_resource(self):
        self.create_flavor()
        self.my_flavor.resource_id = 'flavor_test_id'
        self.my_flavor.client = mock.MagicMock()
        flavors = mock.MagicMock()
        flavor = mock.MagicMock()
        flavor.to_dict.return_value = {'flavor': 'info'}
        flavors.get.return_value = flavor
        self.my_flavor.client().flavors = flavors
        self.assertEqual({'flavor': 'info'}, self.my_flavor.FnGetAtt('show'))
        flavors.get.assert_called_once_with('flavor_test_id')

    def test_flavor_get_live_state(self):
        self.create_flavor()
        value = mock.MagicMock()
        value.get_keys.return_value = {'key': 'value'}
        value.to_dict.return_value = {'ram': 1024,
                                      'disk': 0,
                                      'vcpus': 1,
                                      'rxtx_factor': 1.0,
                                      'OS-FLV-EXT-DATA:ephemeral': 0,
                                      'os-flavor-access:is_public': True}

        self.flavors.get.return_value = value
        self.my_flavor.resource_id = '1234'

        reality = self.my_flavor.get_live_state(self.my_flavor.properties)
        self.assertEqual({'extra_specs': {'key': 'value'}}, reality)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import versionutils
from oslo_policy import policy

from heat.policies import base

# These policies are for AWS CloudFormation-like APIs, so we won't list out
# the URI paths in rules.

DEPRECATED_REASON = """
The cloud formation API now supports system scope and default roles.
"""

POLICY_ROOT = 'cloudformation:%s'

deprecated_list_stacks = policy.DeprecatedRule(
    name=POLICY_ROOT % 'ListStacks',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_create_stack = policy.DeprecatedRule(
    name=POLICY_ROOT % 'CreateStack',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_describe_stacks = policy.DeprecatedRule(
    name=POLICY_ROOT % 'DescribeStacks',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_delete_stack = policy.DeprecatedRule(
    name=POLICY_ROOT % 'DeleteStack',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_update_stack = policy.DeprecatedRule(
    name=POLICY_ROOT % 'UpdateStack',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_cancel_update_stack = policy.DeprecatedRule(
    name=POLICY_ROOT % 'CancelUpdateStack',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_describe_stack_events = policy.DeprecatedRule(
    name=POLICY_ROOT % 'DescribeStackEvents',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_validate_template = policy.DeprecatedRule(
    name=POLICY_ROOT % 'ValidateTemplate',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_get_template = policy.DeprecatedRule(
    name=POLICY_ROOT % 'GetTemplate',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_estimate_template_cost = policy.DeprecatedRule(
    name=POLICY_ROOT % 'EstimateTemplateCost',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_describe_stack_resource = policy.DeprecatedRule(
    name=POLICY_ROOT % 'DescribeStackResource',
    check_str=base.RULE_ALLOW_EVERYBODY,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_describe_stack_resources = policy.DeprecatedRule(
    name=POLICY_ROOT % 'DescribeStackResources',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)
deprecated_list_stack_resources = policy.DeprecatedRule(
    name=POLICY_ROOT % 'ListStackResources',
    check_str=base.RULE_DENY_STACK_USER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY
)

cloudformation_policies = [
    policy.RuleDefault(
        name=POLICY_ROOT % 'ListStacks',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        deprecated_rule=deprecated_list_stacks
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'CreateStack',
        check_str=base.PROJECT_MEMBER,
        scope_types=['project'],
        deprecated_rule=deprecated_create_stack
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'DescribeStacks',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        deprecated_rule=deprecated_describe_stacks
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'DeleteStack',
        check_str=base.PROJECT_MEMBER,
        scope_types=['project'],
        deprecated_rule=deprecated_delete_stack
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'UpdateStack',
        check_str=base.PROJECT_MEMBER,
        scope_types=['project'],
        deprecated_rule=deprecated_update_stack
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'CancelUpdateStack',
        check_str=base.PROJECT_MEMBER,
        scope_types=['project'],
        deprecated_rule=deprecated_cancel_update_stack
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'DescribeStackEvents',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        deprecated_rule=deprecated_describe_stack_events
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'ValidateTemplate',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        deprecated_rule=deprecated_validate_template
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'GetTemplate',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        deprecated_rule=deprecated_get_template
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'EstimateTemplateCost',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        deprecated_rule=deprecated_estimate_template_cost
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'DescribeStackResource',
        check_str=base.PROJECT_READER_OR_STACK_USER,
        scope_types=['project'],
        deprecated_rule=deprecated_describe_stack_resource
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'DescribeStackResources',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        deprecated_rule=deprecated_describe_stack_resources
    ),
    policy.RuleDefault(
        name=POLICY_ROOT % 'ListStackResources',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        deprecated_rule=deprecated_list_stack_resources
    )
]


def list_rules():
    return cloudformation_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools

from heat.common import template_format
from heat.engine.clients.os import glance
from heat.engine.clients.os import keystone
from heat.engine.clients.os.keystone import fake_keystoneclient as fake_ks
from heat.engine.clients.os import nova
from heat.engine import environment
from heat.engine.resources.aws.ec2 import instance as instances
from heat.engine import stack as parser
from heat.engine import template as templatem
from heat.tests.openstack.nova import fakes as fakes_nova
from heat.tests import utils

wp_template = '''
heat_template_version: 2014-10-16
description: WordPress
parameters:
  KeyName:
    description: KeyName
    type: string
    default: test\u2042
resources:
  WebServer:
    type: AWS::EC2::Instance
    properties:
      ImageId: F17-x86_64-gold
      InstanceType: m1.large
      KeyName: test
      UserData: wordpress
'''

string_template_five = '''
heat_template_version: 2013-05-23
description: Random String templates

parameters:
    salt:
        type: string
        default: "quickbrownfox"

resources:
    A:
        type: OS::Heat::RandomString
        properties:
            salt: {get_param: salt}

    B:
        type: OS::Heat::RandomString
        properties:
            salt: {get_param: salt}

    C:
        type: OS::Heat::RandomString
        depends_on: [A, B]
        properties:
            salt: {get_attr: [A, value]}

    D:
        type: OS::Heat::RandomString
        depends_on: C
        properties:
            salt: {get_param: salt}

    E:
        type: OS::Heat::RandomString
        depends_on: C
        properties:
            salt: {get_param: salt}
'''

string_template_five_update = '''
heat_template_version: 2013-05-23
description: Random String templates

parameters:
    salt:
        type: string
        default: "quickbrownfox123"

resources:
    A:
        type: OS::Heat::RandomString
        properties:
            salt: {get_param: salt}

    B:
        type: OS::Heat::RandomString
        properties:
            salt: {get_param: salt}

    F:
        type: OS::Heat::RandomString
        depends_on: [A, B]
        properties:
            salt: {get_param: salt}

    G:
        type: OS::Heat::RandomString
        depends_on: F
        properties:
            salt: {get_param: salt}

    H:
        type: OS::Heat::RandomString
        depends_on: F
        properties:
            salt: {get_param: salt}
'''

attr_cache_template = '''
heat_template_version: 2016-04-08
resources:
    A:
        type: ResourceWithComplexAttributesType
    B:
        type: OS::Heat::RandomString
        properties:
            salt: {get_attr: [A, flat_dict, key2]}
    C:
        type: OS::Heat::RandomString
        depends_on: [A, B]
        properties:
            salt: {get_attr: [A, nested_dict, dict, a]}
    D:
        type: OS::Heat::RandomString
        depends_on: C
        properties:
            salt: {get_attr: [A, nested_dict, dict, b]}
    E:
        type: OS::Heat::RandomString
        depends_on: C
        properties:
            salt: {get_attr: [A, flat_dict, key3]}
'''


def get_stack(stack_name, ctx, template=None, with_params=True,
              convergence=False, **kwargs):
    if template is None:
        t = template_format.parse(wp_template)
        if with_params:
            env = environment.Environment({'KeyName': 'test'})
            tmpl = templatem.Template(t, env=env)
        else:
            tmpl = templatem.Template(t)
    else:
        t = template_format.parse(template)
        tmpl = templatem.Template(t)
    stack = parser.Stack(ctx, stack_name, tmpl, convergence=convergence,
                         **kwargs)
    stack.thread_group_mgr = DummyThreadGroupManager()
    return stack


def setup_keystone_mocks_with_mock(test_case, stack):
    fkc = fake_ks.FakeKeystoneClient()

    test_case.patchobject(keystone.KeystoneClientPlugin, '_create')
    keystone.KeystoneClientPlugin._create.return_value = fkc


def setup_mock_for_image_constraint_with_mock(test_case, imageId_input,
                                              imageId_output=744):
    test_case.patchobject(glance.GlanceClientPlugin,
                          'find_image_by_name_or_id',
                          return_value=imageId_output)


def validate_setup_mocks_with_mock(stack, fc, mock_image_constraint=True,
                                   validate_create=True):
    instance = stack['WebServer']
    metadata = instance.metadata_get()
    if mock_image_constraint:
        m_image = glance.GlanceClientPlugin.find_image_by_name_or_id
        m_image.assert_called_with(
            instance.properties['ImageId'])

    user_data = instance.properties['UserData']
    server_userdata = instance.client_plugin().build_userdata(
        metadata, user_data, 'ec2-user')
    nova.NovaClientPlugin.build_userdata.assert_called_with(
        metadata, user_data, 'ec2-user')

    if not validate_create:
        return

    fc.servers.create.assert_called_once_with(
        image=744,
        flavor=3,
        key_name='test',
        name=utils.PhysName(stack.name, 'WebServer'),
        security_groups=None,
        userdata=server_userdata,
        scheduler_hints=None,
        meta=None,
        nics=None,
        availability_zone=None,
        block_device_mapping=None)


def setup_mocks_with_mock(testcase, stack, mock_image_constraint=True,
                          mock_keystone=True):
    fc = fakes_nova.FakeClient()
    testcase.patchobject(instances.Instance, 'client', return_value=fc)
    testcase.patchobject(nova.NovaClientPlugin, 'client', return_value=fc)
    instance = stack['WebServer']
    metadata = instance.metadata_get()
    if mock_image_constraint:
        setup_mock_for_image_constraint_with_mock(
            testcase, instance.properties['ImageId'])

    if mock_keystone:
        setup_keystone_mocks_with_mock(testcase, stack)

    user_data = instance.properties['UserData']
    server_userdata = instance.client_plugin().build_userdata(
        metadata, user_data, 'ec2-user')
    testcase.patchobject(nova.NovaClientPlugin, 'build_userdata',
                         return_value=server_userdata)

    testcase.patchobject(fc.servers, 'create')

    fc.servers.create.return_value = fc.servers.list()[4]
    return fc


def setup_stack_with_mock(test_case, stack_name, ctx, create_res=True,
                          convergence=False):
    stack = get_stack(stack_name, ctx, convergence=convergence)
    stack.store()
    if create_res:
        fc = setup_mocks_with_mock(test_case, stack)
        stack.create()
        stack._persist_state()
        validate_setup_mocks_with_mock(stack, fc)
    return stack


def clean_up_stack(test_case, stack, delete_res=True):
    if delete_res:
        fc = fakes_nova.FakeClient()
        test_case.patchobject(instances.Instance, 'client', return_value=fc)
        test_case.patchobject(fc.servers, 'delete',
                              side_effect=fakes_nova.fake_exception())
    stack.delete()


def stack_context(stack_name, create_res=True, convergence=False):
    """Decorator for creating and deleting stack.

    Decorator which creates a stack by using the test case's context and
    deletes it afterwards to ensure tests clean up their stacks regardless
    of test success/failure.
    """
    def stack_delete(test_fn):
        @functools.wraps(test_fn)
        def wrapped_test(test_case, *args, **kwargs):
            def create_stack():
                ctx = getattr(test_case, 'ctx', None)
                if ctx is not None:
                    stack = setup_stack_with_mock(test_case, stack_name, ctx,
                                                  create_res, convergence)
                    setattr(test_case, 'stack', stack)

            def delete_stack():
                stack = getattr(test_case, 'stack', None)
                if stack is not None and stack.id is not None:
                    clean_up_stack(test_case, stack, delete_res=create_res)

            create_stack()
            try:
                test_fn(test_case, *args, **kwargs)
            except Exception as err:
                try:
                    delete_stack()
                finally:
                    raise err from None
            else:
                delete_stack()

        return wrapped_test
    return stack_delete


class DummyThread(object):

    def link(self, callback, *args):
        pass


class DummyThreadGroup(object):
    def __init__(self):
        self.threads = []

    def add_timer(self, interval, callback, initial_delay=None,
                  *args, **kwargs):
        self.threads.append(callback)

    def stop_timers(self):
        pass

    def add_thread(self, callback, cnxt, trace, func, *args, **kwargs):
        # callback here is _start_with_trace(); func is the 'real' callback
        self.threads.append(func)
        return DummyThread()

    def stop(self, graceful=False):
        pass

    def wait(self):
        pass


class DummyThreadGroupManager(object):
    def __init__(self):
        self.msg_queues = []
        self.messages = []

    def start(self, stack, func, *args, **kwargs):
        # Just run the function, so we know it's completed in the test
        func(*args, **kwargs)
        return DummyThread()

    def start_with_lock(self, cnxt, stack, engine_id, func, *args, **kwargs):
        # Just run the function, so we know it's completed in the test
        func(*args, **kwargs)
        return DummyThread()

    def start_with_acquired_lock(self, stack, lock, func, *args, **kwargs):
        # Just run the function, so we know it's completed in the test
        func(*args, **kwargs)
        return DummyThread()

    def send(self, stack_id, message):
        self.messages.append(message)

    def add_msg_queue(self, stack_id, msg_queue):
        self.msg_queues.append(msg_queue)

    def remove_msg_queue(self, gt, stack_id, msg_queue):
        for q in self.msg_queues.pop(stack_id, []):
            if q is not msg_queue:
                self.add_event(stack_id, q)


class DummyThreadGroupMgrLogStart(DummyThreadGroupManager):
    def __init__(self):
        super(DummyThreadGroupMgrLogStart, self).__init__()
        self.started = []

    def start_with_lock(self, cnxt, stack, engine_id, func, *args, **kwargs):
        self.started.append((stack.id, func))
        return DummyThread()

    def start_with_acquired_lock(self, stack, lock, func, *args, **kwargs):
        self.started.append((stack.id, func))
        return DummyThread()

    def start(self, stack_id, func, *args, **kwargs):
        # Here we only store the started task so it can be checked
        self.started.append((stack_id, func))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints as constr
from heat.engine import properties
from heat.engine import resource
from heat.engine.resources.openstack.heat import software_config as sc
from heat.engine import support
from heat.rpc import api as rpc_api


class SoftwareComponent(sc.SoftwareConfig):
    """A resource for describing and storing a software component.

    This resource is similar to OS::Heat::SoftwareConfig. In contrast to
    SoftwareConfig which allows for storing only one configuration (e.g. one
    script), SoftwareComponent allows for storing multiple configurations to
    address handling of all lifecycle hooks (CREATE, UPDATE, SUSPEND, RESUME,
    DELETE) for a software component in one place.

    This resource is backed by the persistence layer and the API of the
    SoftwareConfig resource, and only adds handling for the additional
    'configs' property and attribute.
    """

    support_status = support.SupportStatus(version='2014.2')

    PROPERTIES = (
        CONFIGS,
        INPUTS, OUTPUTS,
        OPTIONS,
    ) = (
        'configs',
        sc.SoftwareConfig.INPUTS, sc.SoftwareConfig.OUTPUTS,
        sc.SoftwareConfig.OPTIONS,
    )

    CONFIG_PROPERTIES = (
        CONFIG_ACTIONS, CONFIG_CONFIG, CONFIG_TOOL,
    ) = (
        'actions', 'config', 'tool',
    )

    ATTRIBUTES = (
        CONFIGS_ATTR,
    ) = (
        'configs',
    )

    # properties schema for one entry in the 'configs' list
    config_schema = properties.Schema(
        properties.Schema.MAP,
        schema={
            CONFIG_ACTIONS: properties.Schema(
                # Note: This properties schema allows for custom actions to be
                # specified, which will however require special handling in
                # in-instance hooks. By default, only the standard actions
                # stated below will be handled.
                properties.Schema.LIST,
                _('Lifecycle actions to which the configuration applies. '
                  'The string values provided for this property can include '
                  'the standard resource actions CREATE, DELETE, UPDATE, '
                  'SUSPEND and RESUME supported by Heat.'),
                default=[resource.Resource.CREATE, resource.Resource.UPDATE],
                schema=properties.Schema(properties.Schema.STRING),
                constraints=[
                    constr.Length(min=1),
                ]
            ),
            CONFIG_CONFIG: sc.SoftwareConfig.properties_schema[
                sc.SoftwareConfig.CONFIG
            ],
            CONFIG_TOOL: properties.Schema(
                properties.Schema.STRING,
                _('The configuration tool used to actually apply the '
                  'configuration on a server. This string property has '
                  'to be understood by in-instance tools running inside '
                  'deployed servers.'),
                required=True
            )
        }
    )

    properties_schema = {
        CONFIGS: properties.Schema(
            properties.Schema.LIST,
            _('The list of configurations for the different lifecycle actions '
              'of the represented software component.'),
            schema=config_schema,
            constraints=[constr.Length(min=1)],
            required=True
        ),
        INPUTS: sc.SoftwareConfig.properties_schema[
            sc.SoftwareConfig.INPUTS],
        OUTPUTS: sc.SoftwareConfig.properties_schema[
            sc.SoftwareConfig.OUTPUTS],
        OPTIONS: sc.SoftwareConfig.properties_schema[
            sc.SoftwareConfig.OPTIONS],
    }

    def handle_create(self):
        props = dict(self.properties)
        props[rpc_api.SOFTWARE_CONFIG_NAME] = self.physical_resource_name()
        # use config property of SoftwareConfig to store configs list
        configs = props.pop(self.CONFIGS)
        props[rpc_api.SOFTWARE_CONFIG_CONFIG] = {self.CONFIGS: configs}
        # set 'group' to enable component processing by in-instance hook
        props[rpc_api.SOFTWARE_CONFIG_GROUP] = 'component'

        sc = self.rpc_client().create_software_config(self.context, **props)
        self.resource_id_set(sc[rpc_api.SOFTWARE_CONFIG_ID])

    def _resolve_attribute(self, name):
        """Retrieve attributes of the SoftwareComponent resource.

        'configs' returns the list of configurations for the software
        component's lifecycle actions. If the attribute does not exist,
        an empty list is returned.
        """
        if name == self.CONFIGS_ATTR and self.resource_id:
            with self.rpc_client().ignore_error_by_name('NotFound'):
                sc = self.rpc_client().show_software_config(
                    self.context, self.resource_id)
                # configs list is stored in 'config' property of parent class
                # (see handle_create)
                return sc[rpc_api.SOFTWARE_CONFIG_CONFIG].get(self.CONFIGS)

    def validate(self):
        """Validate SoftwareComponent properties consistency."""
        super(SoftwareComponent, self).validate()

        # One lifecycle action (e.g. CREATE) can only be associated with one
        # config; otherwise a way to define ordering would be required.
        configs = self.properties[self.CONFIGS]
        config_actions = set()
        for config in configs:
            actions = config.get(self.CONFIG_ACTIONS)
            if any(action in config_actions for action in actions):
                msg = _('Defining more than one configuration for the same '
                        'action in SoftwareComponent "%s" is not allowed.'
                        ) % self.name
                raise exception.StackValidationFailed(message=msg)
            config_actions.update(actions)


def resource_mapping():
    return {
        'OS::Heat::SoftwareComponent': SoftwareComponent,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


def exact_filter(query, model, filters):
    """Applies exact match filtering to a query.

    Returns the updated query.  Modifies filters argument to remove
    filters consumed.

    :param query: query to apply filters to
    :param model: model object the query applies to, for IN-style
                  filtering
    :param filters: dictionary of filters; values that are lists,
                    tuples, sets, or frozensets cause an 'IN' test to
                    be performed, while exact matching ('==' operator)
                    is used for other values
    """

    filter_dict = {}
    if filters is None:
        filters = {}

    for key, value in filters.items():
        if isinstance(value, (list, tuple, set, frozenset)):
            column_attr = getattr(model, key)
            query = query.filter(column_attr.in_(value))
        else:
            filter_dict[key] = value

    if filter_dict:
        query = query.filter_by(**filter_dict)

    return query



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import copy
import json
import time

from heat_integrationtests.common import test
from heat_integrationtests.functional import functional_base

_test_template = {
    'heat_template_version': 'pike',
    'description': 'Test template to create two resources.',
    'resources': {
        'test1': {
            'type': 'OS::Heat::TestResource',
            'properties': {
                'value': 'Test1',
                'fail': False,
                'update_replace': False,
                'wait_secs': 0,
            }
        },
        'test2': {
            'type': 'OS::Heat::TestResource',
            'properties': {
                'value': 'Test1',
                'fail': False,
                'update_replace': False,
                'wait_secs': 0,
                'action_wait_secs': {
                    'create': 30,
                }
            },
            'depends_on': ['test1']
        }
    }
}


def get_templates(fail=False, delay_s=None):
    before = copy.deepcopy(_test_template)

    after = copy.deepcopy(before)
    for r in after['resources'].values():
        r['properties']['value'] = 'Test2'

    before_props = before['resources']['test2']['properties']
    before_props['fail'] = fail
    if delay_s is not None:
        before_props['action_wait_secs']['create'] = delay_s

    return before, after


class SimultaneousUpdateStackTest(functional_base.FunctionalTestsBase):

    @test.requires_convergence
    def test_retrigger_success(self):
        before, after = get_templates()
        stack_id = self.stack_create(template=before,
                                     expected_status='CREATE_IN_PROGRESS')
        time.sleep(10)

        self.update_stack(stack_id, after)

    @test.requires_convergence
    def test_retrigger_failure(self):
        before, after = get_templates(fail=True)
        stack_id = self.stack_create(template=before,
                                     expected_status='CREATE_IN_PROGRESS')
        time.sleep(10)

        self.update_stack(stack_id, after)

    @test.requires_convergence
    def test_retrigger_timeout(self):
        before, after = get_templates(delay_s=70)
        stack_id = self.stack_create(template=before,
                                     expected_status='CREATE_IN_PROGRESS',
                                     timeout=1)
        time.sleep(50)

        self.update_stack(stack_id, after)


input_param = 'input'
preempt_nested_stack_type = 'preempt.yaml'
preempt_root_rsrcs = {
    'nested_stack': {
        'type': preempt_nested_stack_type,
        'properties': {
            'input': {'get_param': input_param},
        },
    }
}
preempt_root_out = {'get_attr': ['nested_stack', 'delay_stack']}
preempt_delay_stack_type = 'delay.yaml'
preempt_nested_rsrcs = {
    'delay_stack': {
        'type': preempt_delay_stack_type,
        'properties': {
            'input': {'get_param': input_param},
        },
    }
}
preempt_nested_out = {'get_resource': 'delay_stack'}
preempt_delay_rsrcs = {
    'delay_resource': {
        'type': 'OS::Heat::TestResource',
        'properties': {
            'action_wait_secs': {
                'update': 6000,
            },
            'value': {'get_param': input_param},
        },
    }
}


def _tmpl_with_rsrcs(rsrcs, output_value=None):
    tmpl = {
        'heat_template_version': 'queens',
        'parameters': {
            input_param: {
                'type': 'string',
            },
        },
        'resources': rsrcs,
    }
    if output_value is not None:
        outputs = {'delay_stack': {'value': output_value}}
        tmpl['outputs'] = outputs
    return json.dumps(tmpl)


class SimultaneousUpdateNestedStackTest(functional_base.FunctionalTestsBase):
    @test.requires_convergence
    def test_nested_preemption(self):
        root_tmpl = _tmpl_with_rsrcs(preempt_root_rsrcs,
                                     preempt_root_out)
        files = {
            preempt_nested_stack_type: _tmpl_with_rsrcs(preempt_nested_rsrcs,
                                                        preempt_nested_out),
            preempt_delay_stack_type: _tmpl_with_rsrcs(preempt_delay_rsrcs),
        }
        stack_id = self.stack_create(template=root_tmpl, files=files,
                                     parameters={input_param: 'foo'})
        delay_stack_uuid = self.get_stack_output(stack_id, 'delay_stack')

        # Start an update that includes a long delay in the second nested stack
        self.update_stack(stack_id, template=root_tmpl, files=files,
                          parameters={input_param: 'bar'},
                          expected_status='UPDATE_IN_PROGRESS')
        self._wait_for_resource_status(delay_stack_uuid, 'delay_resource',
                                       'UPDATE_IN_PROGRESS')

        # Update again to check that we preempt update of the first nested
        # stack. This will delete the second nested stack, after preempting the
        # update of that stack as well, which will cause the delay resource
        # within to be cancelled.
        empty_nest_files = {
            preempt_nested_stack_type: _tmpl_with_rsrcs({}),
        }
        self.update_stack(stack_id, template=root_tmpl, files=empty_nest_files,
                          parameters={input_param: 'baz'})



#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""CLI interface for heat management."""

import sys

from oslo_config import cfg
from oslo_log import log

from heat.common import context
from heat.common import exception
from heat.common.i18n import _
from heat.common import messaging
from heat.common import service_utils
from heat.db import api as db_api
from heat.db import migration as db_migration
from heat.objects import service as service_objects
from heat.rpc import client as rpc_client
from heat import version

CONF = cfg.CONF


def do_db_version():
    """Print database's current migration level."""
    print(db_migration.db_version())


def do_db_sync():
    """Place a database under migration control and upgrade.

    Creating first if necessary.
    """
    db_migration.db_sync(CONF.command.version)


class ServiceManageCommand(object):
    def service_list(self):
        ctxt = context.get_admin_context()
        services = [service_utils.format_service(service)
                    for service in service_objects.Service.get_all(ctxt)]

        print_format = "%-16s %-16s %-36s %-10s %-10s %-10s %-10s"
        print(print_format % (_('Hostname'),
                              _('Binary'),
                              _('Engine_Id'),
                              _('Host'),
                              _('Topic'),
                              _('Status'),
                              _('Updated At')))

        for svc in services:
            print(print_format % (svc['hostname'],
                                  svc['binary'],
                                  svc['engine_id'],
                                  svc['host'],
                                  svc['topic'],
                                  svc['status'],
                                  svc['updated_at']))

    def service_clean(self):
        ctxt = context.get_admin_context()
        for service in service_objects.Service.get_all(ctxt):
            svc = service_utils.format_service(service)
            if svc['status'] == 'down':
                service_objects.Service.delete(ctxt, svc['id'])
        print(_('Dead engines are removed.'))

    @staticmethod
    def add_service_parsers(subparsers):
        service_parser = subparsers.add_parser('service')
        service_parser.set_defaults(command_object=ServiceManageCommand)
        service_subparsers = service_parser.add_subparsers(dest='action')
        list_parser = service_subparsers.add_parser('list')
        list_parser.set_defaults(func=ServiceManageCommand().service_list)
        remove_parser = service_subparsers.add_parser('clean')
        remove_parser.set_defaults(func=ServiceManageCommand().service_clean)


def do_resource_data_list():
    ctxt = context.get_admin_context()
    data = db_api.resource_data_get_all(ctxt, CONF.command.resource_id)

    print_format = "%-16s %-64s"

    for k in data.keys():
        print(print_format % (k, data[k]))


def do_reset_stack_status():
    print(_("Warning: this command is potentially destructive and only "
            "intended to recover from specific crashes."))
    print(_("It is advised to shutdown all Heat engines beforehand."))
    print(_("Continue ? [y/N]"))
    data = input()
    if not data.lower().startswith('y'):
        return
    ctxt = context.get_admin_context()
    db_api.reset_stack_status(ctxt, CONF.command.stack_id)


def do_migrate():
    messaging.setup()
    client = rpc_client.EngineClient()
    ctxt = context.get_admin_context()
    try:
        client.migrate_convergence_1(ctxt, CONF.command.stack_id)
    except exception.NotFound:
        raise Exception(_("Stack with id %s can not be found.")
                        % CONF.command.stack_id)
    except (exception.NotSupported, exception.ActionNotComplete) as ex:
        raise Exception(ex.message)


def purge_deleted():
    """Remove database records that have been previously soft deleted."""
    db_api.purge_deleted(CONF.command.age,
                         CONF.command.granularity,
                         CONF.command.project_id,
                         CONF.command.batch_size)


def do_crypt_parameters_and_properties():
    """Encrypt/decrypt hidden parameters and resource properties data."""
    ctxt = context.get_admin_context()
    prev_encryption_key = CONF.command.previous_encryption_key
    if CONF.command.crypt_operation == "encrypt":
        db_api.encrypt_parameters_and_properties(
            ctxt, prev_encryption_key, CONF.command.verbose_update_params)
    elif CONF.command.crypt_operation == "decrypt":
        db_api.decrypt_parameters_and_properties(
            ctxt, prev_encryption_key, CONF.command.verbose_update_params)


def do_properties_data_migrate():
    print(
        'This command has been deprecated and is now a no-op. '
        'It will be removed in a future release.'
    )


def add_command_parsers(subparsers):
    # db_version parser
    parser = subparsers.add_parser('db_version')
    parser.set_defaults(func=do_db_version)

    # db_sync parser
    parser = subparsers.add_parser('db_sync')
    parser.set_defaults(func=do_db_sync)
    # positional parameter, can be skipped. default=None
    parser.add_argument('version', nargs='?')

    # migrate_convergence_1 parser
    parser = subparsers.add_parser('migrate_convergence_1')
    parser.set_defaults(func=do_migrate)
    parser.add_argument('stack_id')

    # purge_deleted parser
    parser = subparsers.add_parser('purge_deleted')
    parser.set_defaults(func=purge_deleted)
    # positional parameter, can be skipped. default='90'
    parser.add_argument('age', nargs='?', default='90',
                        help=_('How long to preserve deleted data.'))
    # optional parameter, can be skipped. default='days'
    parser.add_argument(
        '-g', '--granularity', default='days',
        choices=['days', 'hours', 'minutes', 'seconds'],
        help=_('Granularity to use for age argument, defaults to days.'))
    # optional parameter, can be skipped.
    parser.add_argument(
        '-p', '--project-id',
        help=_('Project ID to purge deleted stacks.'))
    # optional parameter, can be skipped. default='20'
    parser.add_argument(
        '-b', '--batch_size', default='20',
        help=_('Number of stacks to delete at a time (per transaction). '
               'Note that a single stack may have many DB rows '
               '(events, etc.) associated with it.'))

    # update_params parser
    parser = subparsers.add_parser('update_params')
    parser.set_defaults(func=do_crypt_parameters_and_properties)
    # positional parameter, can't be skipped
    parser.add_argument('crypt_operation',
                        choices=['encrypt', 'decrypt'],
                        help=_('Valid values are encrypt or decrypt. The '
                               'heat-engine processes must be stopped to use '
                               'this.'))
    # positional parameter, can be skipped. default=None
    parser.add_argument('previous_encryption_key',
                        nargs='?',
                        default=None,
                        help=_('Provide old encryption key. New encryption'
                               ' key would be used from config file.'))
    parser.add_argument('--verbose-update-params', action='store_true',
                        help=_('Print an INFO message when processing of each '
                               'raw_template or resource begins or ends'))

    parser = subparsers.add_parser('resource_data_list')
    parser.set_defaults(func=do_resource_data_list)
    parser.add_argument('resource_id',
                        help=_('Stack resource id'))

    parser = subparsers.add_parser('reset_stack_status')
    parser.set_defaults(func=do_reset_stack_status)
    parser.add_argument('stack_id',
                        help=_('Stack id'))

    # migrate properties_data parser
    parser = subparsers.add_parser('migrate_properties_data')
    parser.set_defaults(func=do_properties_data_migrate)

    ServiceManageCommand.add_service_parsers(subparsers)


command_opt = cfg.SubCommandOpt('command',
                                title='Commands',
                                help=_('Show available commands.'),
                                handler=add_command_parsers)


def main():
    log.register_options(CONF)
    log.setup(CONF, "heat-manage")
    CONF.register_cli_opt(command_opt)
    try:
        default_config_files = cfg.find_config_files('heat', 'heat-engine')
        CONF(sys.argv[1:], project='heat', prog='heat-manage',
             version=version.version_info.version_string(),
             default_config_files=default_config_files)
    except RuntimeError as e:
        sys.exit("ERROR: %s" % e)

    try:
        CONF.command.func()
    except Exception as e:
        sys.exit("ERROR: %s" % e)
